# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 17:52:28 2020

@author: willi
"""
import numpy as np
from . import CodonDictionaries
from . import translation_models as models
from . import ODE_Soln
from . import SSA_Soln

SSA_Soln = SSA_Soln.SSA_Soln
ODE_Soln = ODE_Soln.ODE_Soln


import time
import warnings
import os

## Search locally for ssa_cpp
path_to_cpp = ''
path_to_gen = ''
path_to_trna = ''

file_path = os.path.dirname(os.path.realpath(__file__))

for root, dirs, files in os.walk(file_path, topdown=False):
   for branch in dirs:

       if 'ssa_cpp' in branch:
           path_to_cpp = os.path.join(root, branch)

       if 'generalized_cpp' in branch:
           path_to_gen = os.path.join(root, branch)
       if 'trna_ssa' in branch:
           path_to_trna = os.path.join(root, branch)
if path_to_cpp != '':
    try:
        cwd = os.getcwd()
        os.chdir(path_to_cpp)

        #import ssa_translation
        import ssa_translation_lowmem
        os.chdir(cwd)
    except:
        os.chdir(cwd)


if path_to_gen != '':
    try:
        cwd = os.getcwd()

        os.chdir(path_to_gen)
        import ssa_translation_lowmem
        os.chdir(cwd)
    except:
        os.chdir(cwd)

if path_to_trna !='':
    try:
        cwd = os.getcwd()

        os.chdir(path_to_trna)
        import ssa_trna
        os.chdir(cwd)
    except:
        os.chdir(cwd)

try:
    import ssa_translation_lowmem
except:
    pass


class Error(Exception):
    """Base class for exceptions in this module."""
    pass

class CythonMissingError(Error):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class TranslationSolvers():
    '''
    Container class for the solvers
    '''
    def __init__(self, time=None,xi=None):
        self.k = None
        self.k_bind = None
        self.k_term = None
        self.multiframe = False
        self.additional_rxns = {}
        self.probe_locations = None
        self.colors = 1

        try:
            ssa_translation_lowmem.np
            self.cython_available = True
        except:
            self.cython_available = False


        self.default_conditions = {'low_mem':True,
                                   'perturb':[0,0,0],
                                   'leaky_probes':False,
                                   'bins':None,
                                   'kprobe':1,
                                   'footprint':9,
                                   'burnin':0,
                                   'record_stats':False,
                                   'kon':1.,
                                   'koff':1.,
                                   'n_traj':30,
                                   'bursting':False,
                                   'tRNA':False,
                                   'ribosome_preallocation':100,
                                   'dynamic_rib_prealloc':False
                                   }

        self.t = np.linspace(0,1000,1001)
        self.x0 = []
        self._poi = None

    @property
    def protein(self):
        return self._poi

    @protein.setter         #Auto detect amount of colors if protein is set
    def protein(self,newpoi):
        self._poi = newpoi
        try:
            self.colors = self._poi.probe_loc.shape[0]
        except:
            pass


    def solve_ballistic_model(self, ki,ke, nt_seq=None,
                              length=None, poi=None, tag= None):
        '''

        Given a iniation rate and set of elongation rates,
        predict analytically the intensity statistics
        and decorrelation time tau.

        **warning this ignores the effect of collisiosn**

        Parameters
        ----------
        ki : float
            initation rate of ribosomes.
        ke : float
            mean elongation rate along the transcript.
        poi : protein object, optional
            a particular protein object to pull geometry from.
            The default is None.
        tag : list, optional
            list of tag locations if geometry not given.
            The default is None.

        Returns
        -------
        tau_analyticals : list of floats
            analytically sovled decorrelation times tau.
        mean_analyticals : list of floats
            the visible ribosome means on the transcript.
        var_analyticals : list of floats
            variances of ribosomes on the transcript.

        '''

        if nt_seq == None:
            nt_seq = poi.nt_seq
            
        if tag == None:
            if poi ==None:
                tags = []
            else:
                colors = np.where((poi.probe_loc)== 1)[0]
                locs = np.where((poi.probe_loc)== 1)[1]
                tags = []
                for i in range(0,max(colors)+1):
    
                    tags.append(locs[np.where(colors == i)].tolist())
        
        #parse out stop codons
        if nt_seq[-3:].lower() in ['taa', 'tag', 'tga' ,'uga' ,'uaa', 'uag']:
            nt_seq = nt_seq[:-3]
        L = len(nt_seq)

        tau_analyticals = []
        mean_analyticals = []
        var_analyticals = []
        
        
        ke_analytical = L*ke / np.sum(self.__get_ui(nt_seq[:-3]))
        tau_analytical = (L )/ke_analytical  #analytical tau ie autocovariance time
        for tag in tags:


            L = poi.total_length #get the total length of the gene
            Lm = np.mean(tag)  #the mean location of the tag epitopes
            L_after_tag = L - tag[-1]
            L_tag = int((tag[-1] - tag[0]) / 2)

            mean_analytical = ki*tau_analytical * (1.-Lm/float(L)) # mean intensity
            var_analytical = ki*tau_analytical * (1.-Lm/float(L))**2  #var intensity
          
            mean_analyticals.append(mean_analytical)
            var_analyticals.append(var_analytical)

        return tau_analytical,mean_analyticals,var_analyticals


    def invert_ballistic(self,tau_measured, mu_I, poi= None):
        '''

        Given a measured decorrelation time tau from FCS,
        predict the elongation rates and initation rates

        Parameters
        ----------
        tau_measured : float, int, list, ndarray
            measured analytical decorrelation time(s).
        mu_I : float, int, list, ndarray
            average intensitie(s).
        poi : Poi object (optional)
            Protein of interest to pull geometry from.

        Returns
        -------
        kes : list
            analytical average k_elongation(s).
        kis : list
            analytical average k_initations(s).

        '''
        if poi == None:
            poi = self._poi
        #if tag == None:
        colors = np.where((poi.probe_loc)== 1)[0]
        locs = np.where((poi.probe_loc)== 1)[1]

        tags = []

        if isinstance(tau_measured,np.ndarray):
            tau_measured = tau_measured.tolist()

        if not isinstance(tau_measured,list):
            tau_measured = [tau_measured]


        if isinstance(mu_I,np.ndarray):
            mu_I = mu_I.tolist()

        if not isinstance(mu_I,list):
            mu_I = [mu_I]


        for i in range(0,max(colors)+1):

            tags.append(locs[np.where(colors == i)].tolist())

        kes = []
        kis = []

        for i in range(len(tags)):
            tag= tags[i]
            L = poi.total_length #get the total length of the gene
            Lm = np.mean(tag)  #the mean location of the tag epitopes
            L_after_tag = L - tag[-1]
            L_tag = int((tag[-1] - tag[0]) / 2)

            ke_analytical = (L)/tau_measured[i]

            ke = ke_analytical/(L)*np.sum(self.__get_ui(poi.nt_seq[:-3]))

            kes.append(ke)
            ki = (mu_I[i]/len(tag)) / ( (1.-Lm/float(L))*tau_measured[i])
            kis.append(ki)

        return kes, kis



    def __get_ui(self, nt_seq, cdict = None):
        '''
        return the ratio of average gene copy number / sequence codon copy number
        '''

        codon_dict = CodonDictionaries.CodonDictionaries()
        if cdict == None:
            cdict = codon_dict.human_codon_frequency_bias_nakamura
        mean_u =codon_dict.mean_genecopynumber
        ui = []
        for i in range(0, len(nt_seq), 3):
            ui.append(mean_u/ cdict[nt_seq[i:i+3]])
        return ui


    '''
    def solve_ssa_set_conditions(self):

        ssa_conditions = self.default_conditions
        kprobe = self.default_conditions['kprobe']
        kon = self.default_conditions['kon']
        koff = self.default_conditions['koff']

        if kprobe != 1:
            ssa_conditions['leaky_probes'] = True
            leaky_probes = True

        perturb = ssa_conditions['perturb']
        leaky_probes = ssa_conditions['leaky_probes']
        low_memory = ssa_conditions['low_mem']
        record_stats = ssa_conditions['record_stats']
        bins = ssa_conditions['bins']
        n_traj = ssa_conditions['n_traj']


        probe_vec = self.protein.probe_vec.astype(np.int32)
        t = self.t

        k = [self.protein.ki,] + self.protein.kelong + [self.protein.kt,]
        self.__check_rates(k)
        x0 = self.x0


        if self.cython_available:
            if low_memory:
                ssa_obj = self.__solve_ssa_lowmem_combined(
                    k,t,x0,n_traj,ssa_conditions=ssa_conditions,
                    kon=kon, koff=koff, kprobe=kprobe)
            else:

                if record_stats:

                    if leaky_probes == False:
                        # if low_memory:
                        #     ssa_obj = self.__solve_ssa_lowmem(k,t,x0,n_traj, ssa_conditions = ssa_conditions)
                        # else:
                        ssa_obj = self.__solve_ssa(k,t,x0,n_traj,ssa_conditions = ssa_conditions)
                    else:
                        # if low_memory:
                        #     ssa_obj = self.__solve_ssa_lowmem_leaky(k,t,x0,k_probe,n_traj, ssa_conditions = ssa_conditions)
                        # else:
                        ssa_obj = self.__solve_ssa_leaky(k,t,x0,kprobe,n_traj,ssa_conditions = ssa_conditions)

                else:
                    if leaky_probes == False:
                        # if low_memory:
                        #     ssa_obj = self.__solve_ssa_lowmem_nostats(k,t,x0,n_traj, ssa_conditions = ssa_conditions)
                        # else:
                        ssa_obj = self.__solve_ssa_nostats(k,t,x0,n_traj,ssa_conditions = ssa_conditions)
                    else:
                        # if low_memory:
                        #     ssa_obj = self.__solve_ssa_lowmem_leaky_nostats(k,t,x0,k_probe,n_traj, ssa_conditions = ssa_conditions)
                        # else:
                        ssa_obj = self.__solve_ssa_leaky_nostats(k,t,x0,kprobe,n_traj,ssa_conditions = ssa_conditions)

        else:
            ssa_obj = self.__solve_ssa_python(k,t,x0,n_traj,ssa_conditions = ssa_conditions, kon=kon, koff=koff, kprobe=kprobe)




        return ssa_obj
    '''

    def solve_ssa_trna(self,k_index, k_diffusion, k_bind, kelong,
                       k_compl, t,x0=[], k_trna = None,
                       perturb=[0,0,0],leaky_probes=False,
                       kprobe=np.ones(1),probe_vec = None, probe_loc=None,
                       kon=1,koff=1,bursting=False,n_traj=10,
                       connection_mat=None):
        
        self.__check_rates_trna(k_index)

        ssa_conditions = self.default_conditions
        if np.sum(kprobe != 1) !=0:
            ssa_conditions['leaky_probes'] = True
            leaky_probes = True

        ssa_conditions['perturb'] = perturb
        ssa_conditions['leaky_probes'] = leaky_probes


        ssa_conditions['bursting'] = bursting

        provided_probe = False
        try:
            probe_vec[0]
            provided_probe = True
        except:
            pass

        provided_protein = False
        try:
            self.protein.kelong[0]
            provided_protein = True
        except:
            pass

        try:
            k_trna[0]
        except:

            strGeneCopy = CodonDictionaries().human_codon_frequency_bias_nakamura
            strGeneCopy.pop('TAG')
            strGeneCopy.pop('TAA')
            strGeneCopy.pop('TGA')

            k_trna = np.array(list(CodonDictionaries().human_codon_frequency_bias_nakamura.values()))


        if not provided_probe:
            if provided_protein:
                probe_vec = self.protein.probe_vec.astype(np.int32)
                probe_loc = self.protein.probe_loc.astype(np.int32)
            else:
                probe_loc = np.zeros([1,len(kelong)], dtype=np.int32)
                probe_vec = np.zeros([1,len(kelong)], dtype=np.int32)
                warnings.warn('no provided probe vector, using a blank probe'\
                              '. This will result in zero sum intensity.')
        else:
            probe_vec = probe_vec


        ssa_conditions['probe_vec'] = probe_vec
        ssa_conditions['probe_loc'] = probe_loc


        try:
            connection_mat[0]
            ssa_obj = self.__solve_ssa_trna_connection_mat(k_index,k_trna, k_diffusion,k_bind,kelong, k_compl,t,x0,n_traj,ssa_conditions = ssa_conditions)

        except:
            ssa_obj = self.__solve_ssa_trna(k_index,k_trna, k_diffusion,k_bind,kelong, k_compl,t,x0,n_traj,ssa_conditions = ssa_conditions)

        return ssa_obj



    def solve_custom_model(self, model, parameters, kelong, t_array,
                           stoich_lattice, stoich_states,
                           xi_lattice, xi_states, n_traj=10,
                           probe_loc=None, poi=None):



        n_total_reactions = int(stoich_states.shape[0] + stoich_lattice.shape[0])
        length = len(kelong)

        try:
            probe_loc[0]
        except:
            probe_loc = poi.probe_loc


        #wash inputs:
        stoich_lattice = stoich_lattice.astype(np.int32)
        stoich_states = stoich_states.astype(np.int32)
        xi_lattice = xi_lattice.astype(np.int32)
        xi_states = xi_states.astype(np.int32)
        parameters = parameters.astype(np.float)
        t_array = t_array.astype(np.float)
        probe_loc = probe_loc.astype(np.int32)

        Ncolors = probe_loc.shape[0]
        max_rib = int(len(kelong)/9 + 5)
        Nt = len(t_array)

        seeds = np.random.randint(0,0x7fffff, n_traj)
        st = time.time()
        all_results = np.zeros([n_traj,  Nt, max_rib,],dtype=np.int32)
        all_intensities = np.zeros([n_traj,  Nt, Ncolors,],dtype=np.int32)
        all_states = np.zeros([n_traj,  Nt, max(xi_states.shape),],dtype=np.int32)

        for i in range(0,n_traj):
            result = np.zeros([Nt, max_rib,],dtype=np.int32)
            intensity = np.zeros([Nt, Ncolors,],dtype=np.int32)
            states = np.zeros([ Nt, max(xi_states.shape),],dtype=np.int32)
            a = model.run_ssa_cpp(result, intensity, states, stoich_states,
                                  stoich_lattice, parameters,
                                  np.array(kelong,dtype=np.float),  t_array,
                                  xi_lattice, xi_states,
                                  probe_loc,  length,
                                  seeds[i], n_total_reactions)
            all_results[i] = result.reshape(max_rib,Nt).T
            all_intensities[i] = intensity.reshape(Ncolors,Nt).T
            all_states[i] = states.reshape(max(xi_states.shape),Nt).T

        eval_time = time.time()-st



        rib_per_t = np.zeros((n_traj,Nt))


        validind = 0
        riblocs = []
        for i in range(len(all_results)):
            try:


                if np.where(np.sum(all_results[i].T,axis=1)!=0)[0][-1] > validind:

                    validind = np.where(np.sum(all_results[i].T,axis=1)!=0)[0][-1]

            #there was a blank
            except:
                pass
            
        if validind == 0:
            validind = 1
        all_results = all_results[:,:,:validind]

        for i in range(all_results.shape[0]):
            rib_loc = all_results[i,:,:]
            rt = np.count_nonzero(rib_loc.T,axis=0)
            rib_per_t[i,:] = rt

        rib_density_loc, rib_density_count = np.unique(all_results,return_counts=True)
        rib_density_loc = rib_density_loc[1:]
        rib_density_count = rib_density_count[1:]
        rib_density = rib_density_count / np.sum(rib_density_count)

        ssa_obj = SSA_Soln()


        ssa_obj.n_traj = n_traj
        ssa_obj.k = kelong
        # ssa_obj.no_rib_per_mrna = no_ribosomes_per_mrna
        ssa_obj.rib_per_t = rib_per_t
        ssa_obj.rib_mean = np.mean(rib_per_t.T)

        ssa_obj.rib_density = rib_density
        #ssa_obj.rib_means = ribosome_means
        ssa_obj.intensity_vec = all_intensities

        ssa_obj.time = t_array
        ssa_obj.time_rec = t_array
        ssa_obj.ribosome_locations = all_results
        ssa_obj.states = all_states
        ssa_obj.eval_time = eval_time

        return ssa_obj




    def solve_ssa(self, ke, t, ki=.33, kt = 10,
                  x0=[], n_traj=100, bins=None, low_memory=True,
                  perturb=[0,0,0,0], leaky_probes=False, kprobe=np.ones(1),
                  record_stats=False, probe_vec=None, probe_loc=None,
                  kon=1, koff=1, bursting=False, rib_prealloc=200,
                  dynamic_prealloc=False):
        '''

        Solve a stochastic simulation of ribosomes on an mRNA transcript
        given codon dependent rates (TASEP)

        Parameters
        ----------
        ke : list
            list of kelongation rates, can be accessed via a protein object by POI.kelong.
        t : list, arr
            a list/array of times to run the simulation for.
        ki : float, optional
            Initiation Rate. The default is .33 1/time.
        kt : float, optional
            Termination Rate. The default is 10 1/time.
        x0 : numpy array, optional
            1xRibsomes max intial condition. The default is []
            (no intial condition).
        n_traj : int, optional
            Number of trajectories to simulate. The default is 100.
        bins : numpy array, optional
            binning key array, for example [2,2,2,2,2] would bin a 10 length
            ke into 5 bins. The default is None.
        low_memory : Bool, optional
            Use lower memory when simulating, meaning dont keep track of
            ribosome positions over time, just the Intensity. The default is True.
        perturb : list [bool, bool, float], optional
            Apply a pertubation [frap, inhibitor, application time], for
            example [1,0,50] would apply frap at time 50. The default is [0,0,0].
        leaky_probes : bool, optional
            Use leaky probes with a probability kprobe for binding when
            passing each epitope. The default is False.
        kprobe : list of floats, optional
            probabilities for each probe to bind, for example a two color
            system should have [.1,.9] for 10% chance of color 1 binding,
            and 90% of color two binding on each epitope.
            The default is np.ones(1).
        record_stats : bool, optional
            Record stats such as collisions or ribosome dwell times.
            The default is False.
        probe_vec : numpy array, optional
            probe cumsum to convert to Intensities Ncolor x L_transcript.
            If left blank will use POI. The default is None.
        probe_loc : numpy array, optional
            probe locations Ncolor x L_transcript with 1 for epitope
            locations, 0 otherwise. If left blank will use POI.
            The default is None.
        kon : float, optional
            on rate for bursting. The default is 1.
        koff : float, optional
            off rate for bursting. The default is 1.
        bursting : bool, optional
            use a on/off bursting dynamic for initiation. The default is False.
        rib_prealloc : int, optional
            how many ribosomes to consider binding at a time, for a
            transcript the maximum that can bind at a time is L/9.
            The default is 200.
        dynamic_prealloc : bool, optional
            calculate the expeted number of maximum ribosomes and use that
            to preallocate memory. The default is False.

        Returns
        -------
        ssa_obj : SSA_Soln_obj
            returns a container object containing all the stats from
            the simulation, such as intensity.

        '''


        self.__check_rates(ke)

        ssa_conditions = self.default_conditions
        if np.sum(kprobe != 1) !=0:
            ssa_conditions['leaky_probes'] = True
            leaky_probes = True

        ssa_conditions['perturb'] = perturb
        ssa_conditions['leaky_probes'] = leaky_probes
        ssa_conditions['low_mem'] = low_memory
        ssa_conditions['record_stats'] = record_stats
        ssa_conditions['bins'] = bins
        ssa_conditions['bursting'] = bursting
        ssa_conditions['ribosome_preallocation'] = rib_prealloc
        ssa_conditions['dynamic_rib_prealloc'] = dynamic_prealloc

        provided_probe = False

        try:
            probe_vec[0]
            provided_probe = True
        except:
            pass

        provided_protein = False
        try:
            self.protein.kelong[0]
            provided_protein = True

        except:
            pass

        if provided_protein: #parse out stop codons if they are included
            includes_stop = False
            if self.protein.aa_seq[-1] == '*' or self.protein.nt_seq[-1].upper() in ['UAA','TAA','TAG','UAG', 'UGA','TGA']:
                if len(self.protein.aa_seq) == len(ke):
                    includes_stop = True



        if not provided_probe:
            if provided_protein:
                if includes_stop:
                    probe_vec = np.require(self.protein.probe_vec.astype(np.int32)[:,:-1],requirements=['C'])
                    probe_loc = np.require(self.protein.probe_loc.astype(np.int32)[:,:-1],requirements=['C'])
                    ke = ke[:-1]
                else:
                    probe_vec = self.protein.probe_vec.astype(np.int32)
                    probe_loc = self.protein.probe_loc.astype(np.int32)
            else:
                probe_loc = np.zeros([1,len(ke)], dtype=np.int32)
                probe_vec = np.zeros([1,len(ke)], dtype=np.int32)
                warnings.warn('no provided probe vector, using a blank probe'\
                              '. This will result in zero sum intensity.')
                
        else:
            probe_vec = probe_vec


        ssa_conditions['probe_vec'] = probe_vec
        ssa_conditions['probe_loc'] = probe_loc


        if dynamic_prealloc:

            ssa_conditions['ribosome_preallocation'] = int(len(ke)/ssa_conditions['footprint'])+5

        k = [ki,] + ke + [kt,]
        if self.cython_available:
            if low_memory:
                ssa_obj = self.__solve_ssa_lowmem_combined(k,t,x0,n_traj,ssa_conditions = ssa_conditions, kon=kon, koff=koff, kprobe=kprobe)
            else:
                ssa_obj = self.__solve_ssa_full_combined(k,t,x0,n_traj,ssa_conditions = ssa_conditions, kon=kon, koff=koff, kprobe=kprobe)


        else:
            ssa_obj = self.__solve_ssa_python(k,t,x0,n_traj,ssa_conditions = ssa_conditions, kon=kon, koff=koff, kprobe=kprobe)


       #Generate metadata
        colors = ssa_obj.intensity_vec.shape[0]
        st = ssa_obj.start_time
        ft = ssa_obj.time_rec[-1]
        lt = len(ssa_obj.time_rec)
        if bins == None:
            bstr = 0
        else:
            bstr=1
        sid = 'L' + str(len(k)) + 'N' + str(n_traj) + 'T' + str(st) + '_' + str(ft) + '_' + str(lt)
        fstr = '-F' + str(int(self.cython_available)) + str(int(sum(perturb))) + str(int(leaky_probes)) + str(int(low_memory))  + str(int(record_stats))  +  str(int(bstr)) + str(int(bursting))
        cstr = 'C' + str(int(colors))
        nprobe = np.sum(probe_loc,axis=1)
        for value in nprobe:
            cstr = cstr+ 'P' + str(int(value))
        sid = sid + cstr + fstr
        ssa_obj._SSA_Soln__meta['id'] = sid

        return ssa_obj



    def solve_ode(self,k,t,x0,ki,pl,bins=None, corr = False):
        '''
        Solve the system of odes describing the mRNA transcript by their forward rates.

        **WARNING** This cannot take into account collisions, if you want to
        simulate collisions you must use stochastic simulations.
        If you dont know if your system will encounter heavy collisions
        use the following rule of thumbs:
            ki ~ min(kelong)  or ki >=  L_transcript / ( 1/np.sum(1/kelong))

        these correspond to a single slow codon causing a jam or v_in >= v_out
        of the total system.

        Parameters
        ----------
        k : list
            elongation rates [ke...., kt].
        t : numpy array / list
            time vector.
        x0 : numpy array
            x0, 1xL_transcript with 1 for ribosomes occupying that location.
        ki : float
            initation rate.
        pl : numpy array
            probe location array, 0s and 1s.
        bins : numpy array, optional
            binning array. The default is None.
        corr : Bool, optional
            calculate the analytical correlation of intensity,
            **THIS WILL TAKE A LONG TIME**. The default is False.

        Returns
        -------
        ode_soln : ODEsoln OBJ
            ODE solution object.

        '''

        st = time.time()
        m_ode = models.TranslateODE()
        m_ode.N = len(k)
        m_ode.tf = t[-1]
        m_ode.ptimes = len(t)
        m_ode.ke = k
        m_ode.kb = ki
        m_ode.fi = 1
        m_ode.ti = t[0]
        m_ode.binary = pl

        #mu_intensity = m_ode.solve()

        m = models.TranslateCorrs()
        m.N = len(k)
        m.tf = t[-1]
        m.ptimes = len(t)
        m.ke = k
        m.kb = ki
        m.fi = 1
        m.ti = t[0]
        m.xi = x0
        m.binary = pl


        ode_soln = ODE_Soln()

        m.get_autonomous_matrix()
        #m.faster_solve()
        m.get_mean_SS()

        mean_ss = m.get_mean_SS()

        mean_I = m.map_to_fluorescence3(mean_ss)
        m.mu_ss = mean_ss

        var_ss = m.get_var_SS()
        var_I = m.map_to_fluorescence(var_ss)


        if corr:

            m.csolve()

            norm_acc = np.ravel((m.intensity)/var_I)
            intensity_acc = m.intensity
            ode_soln.intensity_acc = m.intensity
            ode_soln.intensity_acc_norm = norm_acc



        ode_soln.mu_state_ss = mean_ss
        ode_soln.var_state_ss = var_ss

        ode_soln.mu_It = m.solve()
        ode_soln.mu_I_ss = mean_I
        ode_soln.var_I_ss = var_I
        ode_soln.time = t
        ode_soln.bins = bins
        ode_soln.N = len(k)
        ode_soln.k = k
        ode_soln.x0 = x0
        ode_soln.fi = 1
        ode_soln.ki = ki
        ode_soln.prob_loc = pl
        ode_soln.solve_time = time.time()-st


        ode_soln.probvec = m.get_fluor_vec()
        if len(k) > 100:
            ode_soln.solve_method = 'expv'
        else:
            ode_soln.solve_method = 'odeint'


        return ode_soln



    '''
    def __solve_ssa(self,k,t,x0,n_traj,ssa_conditions=None):

        seeds = np.random.randint(0, 0x7FFFFFF, n_traj, dtype = np.int32)

        if ssa_conditions == None:
            ssa_conditions = self.default_conditions

        x0 = self.__check_x0(x0)


        rib_vec = []
        solutions = []
        solutionssave = []
        N_rib = 200
        colors = self.colors
        all_results,all_nribs,all_collisions,all_frapresults,all_ribtimes,all_col_points = self.__generate_mats(n_traj,k[0],t,N_rib,colors)
        footprint = ssa_conditions['footprint']
        evf = ssa_conditions['perturb'][0]
        evi = ssa_conditions['perturb'][1]
        intime = ssa_conditions['perturb'][2]
        non_consider_time = ssa_conditions['burnin']


        st = time.time()

        for i in range(n_traj):

            result,ribtimes,frapresult,coltimes,colpointsx,colpointst = self.__generate_vecs(k,t,N_rib,colors)
            nribs = np.array([0],dtype=np.int32)
            kelong = np.array(k[1:-1]).flatten()




            ssa_translation.run_SSA(result, ribtimes, coltimes, colpointsx,colpointst, kelong,frapresult, t, k[0], k[-1], evf, evi, intime, seeds[i],nribs,x0,footprint,200)
            #ssa_translation.run_SSA(result, ribtimes, coltimes, k[1:-1],frapresult, truetime, k[0], k[-1], evf, evi, intime, seeds[i],nribs)

            all_results[i, :] = result.T
            all_frapresults[i,:] = frapresult
            all_ribtimes[i,:] = ribtimes
            all_collisions[i,:] = coltimes
            all_nribs[i,:] = nribs

            endcolrec = np.where(colpointsx == 0)[0][0]

            colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
            all_col_points.append(colpoints.T)



        maxso = 0
        for i in range(n_traj):
            soln = all_results[i, :].reshape((N_rib, len(t)))


            validind = np.where(np.sum(soln,axis=1)!=0)[0]


            if np.max(validind) != N_rib-1:
                validind = np.append(np.where(np.sum(soln,axis=1)!=0)[0],np.max(validind)+1)

            so = soln[(validind,)]
            if so.shape[0] > maxso:
                maxso = so.shape[0]



        for i in range(n_traj):
            soln = all_results[i, :].reshape((N_rib, len(t)))


            validind = tuple([x for x in range(0,maxso)])


            so = soln[(validind,)]

            solutionssave.append(so)
            solutions.append(soln)


        collisions = np.array([[]])
        watched_ribs = []
        for i in range(n_traj):
            totalrib = all_nribs[i]

            if totalrib > all_collisions.shape[1]:
                collisions = np.append(collisions, all_collisions[i][:])
                watched_ribs.append(int(all_collisions.shape[1]))

            else:

                collisions = np.append(collisions, all_collisions[i][:int(totalrib[0])])
                watched_ribs.append(int(totalrib[0]))

        sttime = time.time() - st


        no_ribosomes = np.zeros((n_traj, (len(k)+1)))

        startindex = np.where(t >= non_consider_time)[0][0]

        #all_results = all_results[:,startindex*N_rib:]
        pv = self.protein.probe_vec
        I = np.zeros((colors,len(t), n_traj))

        for n in range(colors):
            for i in range(n_traj):
                traj = all_results[i,:].reshape((N_rib,len(t))).T
                for j in range(len(t)):
                    temp_output = traj[j,:]

                    I[n,j,i] = np.sum(pv[n][temp_output[temp_output>0]-1]  )

        for i in range(len(solutions)):
            for j in range(len(solutions[0][0][startindex:])):
                rib_pos = solutions[i][startindex:, j][np.nonzero(solutions[i][startindex:, j])]

                no_ribosomes[i, rib_pos.astype(int)] += 1
        no_ribosomes = no_ribosomes[:, 1:]

        ribosome_means = np.mean(no_ribosomes, axis=0)
        ribosome_density = ribosome_means/len(k)

        no_ribosomes_per_mrna = np.mean(no_ribosomes)

        ssa_obj = SSA_Soln()
        ssa_obj.no_ribosomes = no_ribosomes
        ssa_obj.n_traj = n_traj
        ssa_obj.k = k
        ssa_obj.no_rib_per_mrna = no_ribosomes_per_mrna
        ssa_obj.rib_density = ribosome_density
        ssa_obj.rib_means = ribosome_means
        ssa_obj.rib_vec = rib_vec
        #ssa_obj.intensity_vec = intensity_vec
        ssa_obj.time_vec_fixed = t
        ssa_obj.time = t
        ssa_obj.time_rec = t[startindex:]
        ssa_obj.start_time = non_consider_time
        ssa_obj.watched_ribs = watched_ribs
        ssa_obj.intensity_vec = I
        ssa_obj.solutions = np.array(solutionssave)
        try:
            ssa_obj.col_points = all_col_points
        except:
            pass



        return ssa_obj

    '''


    def __map_to_intensity(self):

        return 1


    def __solve_ssa_trna(self, kindex, ktrna, kdiffusion, kbind, kelong,
                         kcompl, t, x0, n_traj, ssa_conditions=None, probe_vec=None ):

        seeds = np.random.randint(0, 0x7FFFFFF, n_traj, dtype=np.int32)

        if ssa_conditions == None:
            ssa_conditions = self.default_conditions

        x0 = self.__check_x0(x0)


        rib_vec = []
        solutions = []
        solutionssave = []
        N_rib = 200
        
        if probe_vec == None:
            pv = self.protein.probe_vec
        else:
            pv = probe_vec
            
        colors = int(pv.shape[0])

        n_trajectories = n_traj


        all_results,all_nribs,all_collisions,all_frapresults,all_ribtimes,all_col_points = self.__generate_mats(n_traj,kbind,t,N_rib,colors)
        all_trna_results = np.zeros((n_trajectories,61*len(t)), dtype=np.int32)

        footprint = ssa_conditions['footprint']
        evf = ssa_conditions['perturb'][0]
        evi = ssa_conditions['perturb'][1]
        intime = ssa_conditions['perturb'][2]
        non_consider_time = ssa_conditions['burnin']

        st = time.time()

        N_rib = 200


        result = np.zeros((len(t)*N_rib), dtype=np.int32  )

        #lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t<20)[0]))
        #all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)

        all_ribs = np.zeros((n_trajectories, 1))
        all_col_points = []

        #seeds = np.random.randint(0,0x7FFFFFF,n_trajectories)
        x0 = np.zeros((N_rib),dtype=np.int32)


        for i in range(n_trajectories):



            trna_result = np.zeros((len(t)*61),dtype=np.int32)
            result,ribtimes,frapresult,coltimes,colpointsx,colpointst = self.__generate_vecs_trna(kbind,kindex,t,N_rib,colors)

            nribs = np.array([0],dtype=np.int32)

            if i == 0: #detect any int64
                inputs = [result,trna_result,ribtimes,coltimes,colpointsx,colpointst, kindex,ktrna,kdiffusion,frapresult,t,kbind,kcompl, seeds[i],nribs,x0,kelong]
                wash_inputs = self.__check_input_memview(inputs)

            if wash_inputs:
                #check memview so all given variables are in int32 if integer for C
                inputs = [result,trna_result,ribtimes,coltimes,colpointsx,colpointst, kindex,ktrna,kdiffusion,frapresult,t,kbind,kcompl, seeds[i],nribs,x0,kelong]

                result,trna_result,ribtimes,coltimes,colpointsx,colpointst, kindex,ktrna,kdiffusion,frapresult,t,kbind,kcompl, seed,nribs,x0,kelong = self.__check_memview(inputs)
            else:

                seed = seeds[i]

            ssa_translation_lowmem.run_SSA_trna_full(result,trna_result,ribtimes,coltimes,colpointsx,colpointst, kindex,ktrna,kdiffusion,frapresult,t,kbind,kcompl, 0,0,0, seed,nribs,x0,kelong)

            all_results[i, :] = result.T
            all_trna_results[i, :] = trna_result
           # all_frapresults[i, :] = frapresult
            all_ribtimes[i, :] = ribtimes
            all_collisions[i, :] = coltimes
            all_nribs[i, :] = nribs[0]

            endcolrec = np.where(colpointsx == 0)[0][0]

            colpoints = np.vstack((colpointsx[:endcolrec],
                                   colpointst[:endcolrec]))
            all_col_points.append(colpoints.T)

        maxso = 0
        for i in range(n_traj):
            soln = all_results[i, :].reshape((N_rib, len(t)))


            validind = np.where(np.sum(soln,axis=1)!=0)[0]


            if np.max(validind) != N_rib-1:
                validind = np.append(
                    np.where(np.sum(soln, axis=1)!=0)[0], np.max(validind)+1)

            so = soln[(validind,)]
            if so.shape[0] > maxso:
                maxso = so.shape[0]



        for i in range(n_traj):
            soln = all_results[i, :].reshape((N_rib, len(t)))


            validind = tuple([x for x in range(0,maxso)])
            so = soln[(validind,)]
            solutionssave.append(so)
            solutions.append(soln)

        collisions = np.array([[]])
        watched_ribs = []
        for i in range(n_traj):
            totalrib = all_nribs[i]

            if totalrib > all_collisions.shape[1]:
                collisions = np.append(collisions, all_collisions[i][:])
                watched_ribs.append(int(all_collisions.shape[1]))

            else:

                collisions = np.append(
                    collisions, all_collisions[i][:int(totalrib[0])])
                watched_ribs.append(int(totalrib[0]))

        sttime = time.time() - st


        no_ribosomes = np.zeros((n_traj, (len(kindex)+1)))

        startindex = np.where(t >= non_consider_time)[0][0]

        #all_results = all_results[:,startindex*N_rib:]
        
        I = np.zeros((colors, len(t), n_traj))

        for n in range(colors):
            for i in range(n_traj):
                traj = all_results[i, :].reshape((N_rib, len(t))).T
                for j in range(len(t)):
                    temp_output = traj[j, :]

                    I[n, j, i] = np.sum(pv[n][temp_output[temp_output>0]-1]  )

        for i in range(len(solutions)):
            for j in range(len(solutions[0][0][startindex:])):
                rib_pos = solutions[i][startindex:, j][np.nonzero(solutions[i][startindex:, j])]

                no_ribosomes[i, rib_pos.astype(int)] += 1
        no_ribosomes = no_ribosomes[:, 1:]

        ribosome_means = np.mean(no_ribosomes, axis=0)
        ribosome_density = ribosome_means/len(kindex)

        no_ribosomes_per_mrna = np.mean(no_ribosomes)

        ssa_obj = SSA_Soln()
        ssa_obj.no_ribosomes = no_ribosomes
        ssa_obj.n_traj = n_traj
        ssa_obj.k = kindex
        ssa_obj.no_rib_per_mrna = no_ribosomes_per_mrna
        ssa_obj.rib_density = ribosome_density
        ssa_obj.rib_means = ribosome_means
        ssa_obj.rib_vec = rib_vec
        ssa_obj.I = I
        ssa_obj.eval_time = sttime
        ssa_obj.time_vec_fixed = t
        ssa_obj.time = t
        ssa_obj.time_rec = t[startindex:]
        ssa_obj.start_time = non_consider_time
        ssa_obj.watched_ribs = watched_ribs
        ssa_obj.intensity_vec = I
        ssa_obj.solutions = np.array(solutionssave)
        ssa_obj.all_trna_results = all_trna_results
        ssa_obj.ribtimes = all_ribtimes
        try:
            ssa_obj.col_points = all_col_points
        except:
            pass



        return ssa_obj

    def __solve_ssa_python(self, k, t, x0, n_traj, ssa_conditions=None,
                           kprobe=None, kon=None, koff=None, flags=[0,0,0]):

        seeds = np.random.randint(0, 0x7FFFFFF, n_traj, dtype=np.int32)

        if ssa_conditions == None:
            ssa_conditions = self.default_conditions

        x0 = self.__check_x0(x0)

        flags = [int(ssa_conditions['bursting']),
                 int(ssa_conditions['leaky_probes']),
                 int(ssa_conditions['record_stats'])]

        rib_vec = []
        solutions = []
        solutionssave = []
        N_rib = 200
        colors = self.colors

        all_results,all_nribs,all_collisions,all_frapresults,all_ribtimes,all_col_points = self.__generate_mats(n_traj,k[0],t,N_rib,colors)
        footprint = ssa_conditions['footprint']
        evf = ssa_conditions['perturb'][0]
        evi = ssa_conditions['perturb'][1]
        intime = ssa_conditions['perturb'][2]
        non_consider_time = ssa_conditions['burnin']


        st = time.time()


        warnings.warn('C++ extention missing or not compiled, using Python' \
                      ' Implementation. This will be much slower!' \
                      ' For C++ instillation instructions read the README or '\
                          'visit the github page: https://github.com/MunskyGroup/rSNAPsim')
        rib_vec = []

        solutions = []
        solutionssave = []
        N_rib = ssa_conditions['ribosome_preallocation']
        collisions = np.array([[]])
        all_results = np.zeros((n_traj, N_rib*len(t)), dtype=np.int32)
        I_internal = np.zeros((colors, len(t), n_traj))
        all_col_points = []
        watched_ribs = []
        for i in range(n_traj):

            soln,all_ribtimes,Ncol,col_points,intensity = self.__ssa_python(k, t, inhibit_time=intime+non_consider_time, FRAP=evf, Inhibitor=evi, flags=flags, kon=kon, kprobe=kprobe, koff=koff, ssa_conditions=ssa_conditions)
            #soln = soln.reshape((1, (len(time_vec_fixed)*N_rib)))

            collisions = np.append(collisions,Ncol)
            watched_ribs.append(int(len(collisions)))
            validind = np.where(np.sum(soln,axis=1) != 0)[0]
            all_col_points.append(np.array(col_points))
            if np.max(validind) != N_rib-1:
                validind = np.append(np.where(np.sum(soln, axis=1) != 0)[0],np.max(validind)+1)

            so = soln[(validind,)]

            #solutionssave.append(so)

            solutions.append(soln)

            result = soln.reshape((1, (len(t)*N_rib)))
            all_results[i, :] = result
            I_internal[:, : ,i] = intensity


        maxso = 0
        for i in range(n_traj):
            soln = all_results[i, :].reshape((N_rib, len(t)))

            validind = np.where(np.sum(soln,axis=1)!=0)[0]
            if np.max(validind) != N_rib-1:
                validind = np.append(np.where(np.sum(soln, axis=1)!=0)[0], np.max(validind)+1)

            so = soln[(validind,)]
            if so.shape[0] > maxso:
                maxso = so.shape[0]

        for i in range(n_traj):
            soln = all_results[i, :].reshape((N_rib, len(t)))

            validind = tuple([x for x in range(0, maxso)])
            so = soln[(validind,)]

            solutionssave.append(so)


        # for i in range(n_traj):
        #     # result,ribtimes,frapresult,coltimes,colpointsx,colpointst = self.__generate_vecs(k,t,N_rib,colors)
        #     # nribs = np.array([0],dtype=np.int32)
        #     # kelong = np.array(k[1:-1]).flatten()

        #     soln,all_ribtimes,Ncol,col_points  = self.__ssa_python(k,t,inhibit_time=intime+non_consider_time,FRAP=evf,Inhibitor=evi,flags=flags)

        #     all_results[i, :] = result.T
        #     all_frapresults[i,:] = frapresult
        #     all_ribtimes[i,:] = ribtimes
        #     all_collisions[i,:] = coltimes
        #     all_nribs[i,:] = nribs

        #     endcolrec = np.where(colpointsx == 0)[0][0]

        #     colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
        #     all_col_points.append(colpoints.T)


        # for i in range(n_traj):
        #     soln = all_results[i, :].reshape((N_rib, len(t)))


        #     validind = np.where(np.sum(soln,axis=1)!=0)[0]

        #     if np.max(validind) != N_rib-1:
        #         validind = np.append(np.where(np.sum(soln,axis=1)!=0)[0],np.max(validind)+1)

        #     so = soln[(validind,)]

        #     solutionssave.append(so)
        #     solutions.append(soln)

        collisions = np.array([[]])
        watched_ribs = []
        for i in range(n_traj):
            totalrib = all_nribs[i]

            if totalrib > all_collisions.shape[1]:
                collisions = np.append(collisions, all_collisions[i][:])
                watched_ribs.append(int(all_collisions.shape[1]))

            else:

                collisions = np.append(collisions, all_collisions[i][:int(totalrib[0])])
                watched_ribs.append(int(totalrib[0]))

        sttime = time.time() - st


        no_ribosomes = np.zeros((n_traj, (len(k)+1)))

        startindex = np.where(t >= non_consider_time)[0][0]

        #all_results = all_results[:,startindex*N_rib:]
        pv = self.protein.probe_vec
        # I = np.zeros((colors,len(t), n_traj))

        # for n in range(colors):
        #     for i in range(n_traj):
        #         traj = all_results[i,:].reshape((N_rib,len(t))).T
        #         for j in range(len(t)):
        #             temp_output = traj[j,:]

        #             I[n,j,i] = np.sum(pv[n][temp_output[temp_output>0]-1]  )

        for i in range(len(solutions)):
            for j in range(len(solutions[0][0][startindex:])):

                rib_pos = solutions[i][startindex:, j][np.nonzero(solutions[i][startindex:, j])]
                if len(rib_pos.astype(int)) > 0:
                    no_ribosomes[i, rib_pos.astype(int)] += 1


        no_ribosomes = no_ribosomes[:, 1:]

        ribosome_means = np.mean(no_ribosomes, axis=0)
        ribosome_density = ribosome_means/len(k)

        no_ribosomes_per_mrna = np.mean(no_ribosomes)

        ssa_obj = SSA_Soln()
        ssa_obj.no_ribosomes = no_ribosomes
        ssa_obj.n_traj = n_traj
        ssa_obj.k = k
        ssa_obj.no_rib_per_mrna = no_ribosomes_per_mrna
        ssa_obj.rib_density = ribosome_density
        ssa_obj.rib_means = ribosome_means
        ssa_obj.rib_vec = rib_vec
        #ssa_obj.intensity_vec = intensity_vec
        ssa_obj.time_vec_fixed = t
        ssa_obj.time = t
        ssa_obj.time_rec = t[startindex:]
        ssa_obj.start_time = non_consider_time
        ssa_obj.watched_ribs = watched_ribs
        ssa_obj.intensity_vec = I_internal
        ssa_obj.I = I_internal
        ssa_obj.solutions = np.array(solutionssave)
        try:
            ssa_obj.col_points = all_col_points
        except:
            pass



        return ssa_obj



    def __solve_ssa_lowmem_combined(self, k, t, x0, n_traj,
                                    ssa_conditions=None, kon=1,
                                    koff=1, kprobe=[]):
        seeds = np.random.randint(0, 0x7FFFFFF, n_traj, dtype=np.int32)

        if isinstance(k, list):
            k = np.array(k).astype(np.float64)

        k = k.flatten()

        if kprobe == []:
            kprobe = np.ones(self.color)

        if isinstance(kprobe, list):
            kprobe = np.array(kprobe, dtype=np.float64)
        if isinstance(kprobe, int):
            kprobe = np.array([kprobe], dtype=np.float64)
        if isinstance(kprobe, float):
            kprobe = np.array([kprobe], dtype=np.float64)


        if ssa_conditions == None:
            ssa_conditions = self.default_conditions

        x0 = self.__check_x0(x0)

        probe_vec = ssa_conditions['probe_vec']

        colors = int(probe_vec.shape[0])
 

        if ssa_conditions['bursting'] == False:
            kon = 1
            koff = 1
        else:
            kon = kon
            koff = koff

        flags = np.array([int(ssa_conditions['bursting']),
                          int(ssa_conditions['leaky_probes']),
                          int(ssa_conditions['record_stats'])], dtype=np.int32)
        probe_loc = ssa_conditions['probe_loc']


        rib_vec = []
        solutions = []
        solutionssave = []


        N_rib = ssa_conditions['ribosome_preallocation']
        all_results,all_nribs,all_collisions,_,all_ribtimes,all_col_points = self.__generate_mats_lowmem(n_traj,k[0],t,N_rib,colors)
        footprint = ssa_conditions['footprint']
        evf = ssa_conditions['perturb'][0]
        evi = ssa_conditions['perturb'][1]
        intime = ssa_conditions['perturb'][2]
        intime_stop = ssa_conditions['perturb'][3]
        non_consider_time = ssa_conditions['burnin']

        st = time.time()

        for i in range(n_traj):

            result,ribtimes,_,coltimes,colpointsx,colpointst = self.__generate_vecs_lowmem(k,t,N_rib,colors)
            nribs = np.array([0], dtype=np.int32)


            if i == 0: #detect any int64
                inputs = [result, ribtimes, coltimes, colpointsx,colpointst, k[1:-1], t, k[0], float(k[-1]), int(evf), int(evi), float(intime), float(intime_stop), seeds[i],nribs,x0,footprint, probe_vec ,int(colors), kon, koff, kprobe, probe_loc, flags,N_rib]
                wash_inputs = self.__check_input_memview(inputs)

            if wash_inputs:
                #check memview so all given variables are in int32 if integer for C
                inputs = [result, ribtimes, coltimes, colpointsx,colpointst, k[1:-1], t, k[0], float(k[-1]), int(evf), int(evi), float(intime),float(intime_stop), seeds[i],nribs,x0,footprint, probe_vec ,int(colors), kon, koff, kprobe, probe_loc, flags,N_rib]
                result, ribtimes, coltimes, colpointsx,colpointst, kelong, t, ki, kt, evf, evi, intime,intime_stop, seed,nribs,x0,footprint, probe_vec ,colors, kon, koff, kprobe, probe_loc, flags, N_rib = self.__check_memview(inputs)
            else:
                kelong = k[1:-1]
                ki = k[0]
                kt = float(k[-1])
                evf = int(evf)
                evi = int(evi)
                intime = float(intime)
                intime_stop = float(intime_stop)
                seed = seeds[i]
                colors = int(colors)
           
            ssa_translation_lowmem.run_SSA(result, ribtimes, coltimes, colpointsx,colpointst, kelong, t, ki, kt, evf, evi, intime,intime_stop, seed,nribs,x0,footprint, probe_vec ,colors, kon, koff, kprobe, probe_loc, flags, N_rib)
            #ssa_translation.run_SSA(result, ribtimes, coltimes, k[1:-1],frapresult, truetime, k[0], k[-1], evf, evi, intime, seeds[i],nribs)

            all_results[i, :, :] = result.T
            #all_frapresults[i, :] = frapresult
            all_ribtimes[i, :] = ribtimes
            all_collisions[i, :] = coltimes
            all_nribs[i, :] = nribs

            endcolrec = np.where(colpointsx == 0)[0][0]

            colpoints = np.vstack((colpointsx[:endcolrec], colpointst[:endcolrec]))
            all_col_points.append(colpoints.T)


            # for i in range(n_traj):
            #     soln = all_results[i, :].reshape((N_rib, len(t),colors))

            #     so = soln
            #     solutionssave.append(so)
            #     solutions.append(soln)

            collisions = np.array([[]])
            watched_ribs = []
            for i in range(n_traj):
                totalrib = all_nribs[i]

                if totalrib > all_collisions.shape[1]:
                    collisions = np.append(collisions, all_collisions[i][:])
                    watched_ribs.append(int(all_collisions.shape[1]))

                else:

                    collisions = np.append(collisions, all_collisions[i][:int(totalrib[0])])
                    watched_ribs.append(int(totalrib[0]))

            sttime = time.time() - st


        no_ribosomes = np.zeros((n_traj, (len(k)+1)))

        startindex = np.where(t >= non_consider_time)[0][0]

        #all_results = all_results[:,startindex*N_rib:]

#        for i in range(len(solutions)):
#            for j in range(len(solutions[0][0][startindex:])):
#                rib_pos = solutions[i][startindex:, j][np.nonzero(solutions[i][startindex:, j])]
#                print(rib_pos)
#
#                no_ribosomes[i, rib_pos.astype(int)] += 1
#        no_ribosomes = no_ribosomes[:, 1:]
#
#        ribosome_means = np.mean(no_ribosomes, axis=0)
#        ribosome_density = ribosome_means/len(k)
#
#        no_ribosomes_per_mrna = np.mean(no_ribosomes)

        ssa_obj = SSA_Soln()
        if ssa_conditions['record_stats']:
            ssa_obj.no_ribosomes = no_ribosomes
            ssa_obj.watched_ribs = watched_ribs
            ssa_obj.collisions = collisions

        ssa_obj.n_traj = n_traj
        ssa_obj.k = k
        #ssa_obj.no_rib_per_mrna = no_ribosomes_per_mrna
        #ssa_obj.rib_density = ribosome_density
        #ssa_obj.rib_means = ribosome_means

        ssa_obj.intensity_vec = all_results.T[:,startindex:,:]
        #ssa_obj.I = all_results.T[:,startindex:,:]

        ssa_obj.time = t
        ssa_obj.time_rec = t[startindex:]
        ssa_obj.start_time = non_consider_time
        ssa_obj.start_index = int(startindex)



        try:
            ssa_obj.col_points = all_col_points
        except:
            pass


        ssa_obj.eval_time = float(sttime)

        return ssa_obj


    def __solve_ssa_full_combined(self,k, t, x0, n_traj, ssa_conditions=None,
                                  kon=1, koff=1, kprobe=[] ):
        seeds = np.random.randint(0, 0x7FFFFFF, n_traj, dtype=np.int32)

        if isinstance(k, list):
            k = np.array(k).astype(np.float64)

        k = k.flatten()

        if kprobe == []:
            kprobe = np.ones(self.color)

        if isinstance(kprobe, list):
            kprobe = np.array(kprobe, dtype=np.float64)
        if isinstance(kprobe, int):
            kprobe = np.array([kprobe], dtype=np.float64)
        if isinstance(kprobe, float):
            kprobe = np.array([kprobe], dtype=np.float64)


        if ssa_conditions == None:
            ssa_conditions = self.default_conditions

        x0 = self.__check_x0(x0)

        probe_vec = ssa_conditions['probe_vec']

        colors = self.colors


        if ssa_conditions['bursting'] == False:
            kon = 1
            koff = 1
        else:
            kon = kon
            koff = koff

        flags = np.array([int(ssa_conditions['bursting']),
                          int(ssa_conditions['leaky_probes']),
                          int(ssa_conditions['record_stats'])], dtype=np.int32)
        probe_loc = ssa_conditions['probe_loc']

        colors = int(probe_loc.shape[0])
        rib_vec = []
        solutions = []
        solutionssave = []


        N_rib = ssa_conditions['ribosome_preallocation']
        all_results,all_nribs,all_collisions,_,all_ribtimes,all_col_points = self.__generate_mats_lowmem(n_traj,k[0],t,N_rib,colors)
        all_rib_loc = np.zeros((n_traj, len(t), N_rib), dtype=np.int32)


        footprint = ssa_conditions['footprint']
        evf = ssa_conditions['perturb'][0]
        evi = ssa_conditions['perturb'][1]
        intime = ssa_conditions['perturb'][2]
        intime_stop = ssa_conditions['perturb'][3]
        non_consider_time = ssa_conditions['burnin']

        st = time.time()

        for i in range(n_traj):

            nribs = np.array([0], dtype=np.int32)
            result,ribtimes,_,coltimes,colpointsx,colpointst = self.__generate_vecs_lowmem(k,t,N_rib,colors)
            ribloc = np.zeros((N_rib,len(t)), dtype=np.int32)
            if i == 0: #detect any int64
                inputs = [ribloc,result, ribtimes, coltimes, colpointsx,colpointst, k[1:-1], t, k[0], float(k[-1]), int(evf), int(evi), float(intime), float(intime_stop), seeds[i],nribs,x0,footprint, probe_vec ,int(colors), kon, koff, kprobe, probe_loc, flags,N_rib]
                wash_inputs = self.__check_input_memview(inputs)

            if wash_inputs:
                #check memview so all given variables are in int32 if integer for C
                inputs = [ribloc,result, ribtimes, coltimes, colpointsx,colpointst, k[1:-1], t, k[0], float(k[-1]), int(evf), int(evi), float(intime), float(intime_stop), seeds[i],nribs,x0,footprint, probe_vec ,int(colors), kon, koff, kprobe, probe_loc, flags,N_rib]
                ribloc,result, ribtimes, coltimes, colpointsx,colpointst, kelong, t, ki, kt, evf, evi, intime,intime_stop, seed,nribs,x0,footprint, probe_vec ,colors, kon, koff, kprobe, probe_loc, flags, N_rib = self.__check_memview(inputs)
            else:
                kelong = k[1:-1]
                ki = k[0]
                kt = float(k[-1])
                evf = int(evf)
                evi = int(evi)
                intime = float(intime)
                intime_stop = float(intime_stop)
                seed = seeds[i]
                colors = int(colors)

            
            ssa_translation_lowmem.run_SSA_full(ribloc, result, ribtimes, coltimes, colpointsx,colpointst, kelong, t, ki, kt, evf, evi, intime, intime_stop, seed,nribs,x0,footprint, probe_vec ,colors, kon, koff, kprobe, probe_loc, flags)
            #ssa_translation.run_SSA(result, ribtimes, coltimes, k[1:-1],frapresult, truetime, k[0], k[-1], evf, evi, intime, seeds[i],nribs)

            all_results[i, :, :] = result.T
            #all_frapresults[i, :] = frapresult
            all_ribtimes[i, :] = ribtimes
            all_collisions[i, :] = coltimes
            all_nribs[i, :] = nribs
            all_rib_loc[i, :, :] = ribloc.T

            endcolrec = np.where(colpointsx == 0)[0][0]

            colpoints = np.vstack((colpointsx[:endcolrec],
                                   colpointst[:endcolrec]))
            all_col_points.append(colpoints.T)



            # for i in range(n_traj):
            #     soln = all_results[i,:, :]

            #     so = soln
            #     solutionssave.append(so)
            #     solutions.append(soln)

            collisions = np.array([[]])
            watched_ribs = []
            for i in range(n_traj):
                totalrib = all_nribs[i]

                if totalrib > all_collisions.shape[1]:
                    collisions = np.append(collisions, all_collisions[i][:])
                    watched_ribs.append(int(all_collisions.shape[1]))

                else:

                    collisions = np.append(collisions, all_collisions[i][:int(totalrib[0])])
                    watched_ribs.append(int(totalrib[0]))

            sttime = time.time() - st


        no_ribosomes = np.zeros((n_traj, (len(k)+1)))

        startindex = np.where(t >= non_consider_time)[0][0]

        all_col_points2 = []
        for i in range(len(all_col_points)):
            all_col_points2.append(all_col_points[i][ all_col_points[i][:, 1] > non_consider_time])


        rib_per_t = np.zeros((n_traj, len(t)))


        validind = 0
        riblocs = []
        for i in range(len(all_rib_loc)):
            try:


                if np.where(np.sum(all_rib_loc[i].T, axis=1)!=0)[0][-1] > validind:

                    validind = np.where(np.sum(all_rib_loc[i].T, axis=1)!=0)[0][-1]

                        #there was a blank
            except:
                pass


        all_rib_loc = all_rib_loc[:, :, :validind]

        for i in range(all_rib_loc.shape[0]):
            rib_loc = all_rib_loc[i, :, :]
            rt = np.count_nonzero(rib_loc.T,axis=0)
            rib_per_t[i, :] = rt

        rib_density_loc, rib_density_count = np.unique(all_rib_loc, return_counts=True)
        rib_density_loc = rib_density_loc[1:]
        rib_density_count = rib_density_count[1:]
        rib_density = rib_density_count / np.sum(rib_density_count)


        ssa_obj = SSA_Soln()
        ssa_obj.no_ribosomes = no_ribosomes
        ssa_obj.n_traj = n_traj
        ssa_obj.k = k
        # ssa_obj.no_rib_per_mrna = no_ribosomes_per_mrna
        ssa_obj.rib_per_t = rib_per_t.T[startindex:, :]
        ssa_obj.rib_mean = np.mean(rib_per_t.T[startindex:, :])

        ssa_obj.rib_density = rib_density
        # ssa_obj.rib_means = ribosome_means
        ssa_obj.intensity_vec = all_results.T[:, startindex:, :]

        ssa_obj.time = t
        ssa_obj.time_rec = t[startindex:]
        ssa_obj.start_time = non_consider_time
        ssa_obj.start_index = int(startindex)
        ssa_obj.watched_ribs = watched_ribs
        ssa_obj.collisions = collisions
        ssa_obj.ribosome_locations = all_rib_loc
        ssa_obj.ribtimes = all_ribtimes


        try:
            ssa_obj.col_points = all_col_points2
        except:
            pass


        ssa_obj.eval_time = sttime

        return ssa_obj


    @classmethod
    def __get_ribosome_statistics(self, ssa_obj, result):


        return ssa_obj


    @classmethod
    def __generate_vecs(cls, k, t, N_rib, ncolor):
        tf = t[-1]
        ki = k[0]

        guessed_no_ribosomes = int(1.3*ki*tf)
        result = np.zeros((len(t)*N_rib), dtype=np.int32)
        ribtimes = np.zeros((guessed_no_ribosomes), dtype=np.float64)
        frapresult = np.zeros((len(t)*N_rib), dtype=np.int32)
        coltimes = np.zeros((guessed_no_ribosomes), dtype=np.int32)
        colpointsx = np.zeros(len(k[1:-1])*(guessed_no_ribosomes), dtype=np.int32)
        colpointst = np.zeros(len(k[1:-1])*(guessed_no_ribosomes), dtype=np.float64)
        return result, ribtimes, frapresult, coltimes, colpointsx, colpointst

    @classmethod
    def __generate_vecs_trna(cls, kbind, kind, t, N_rib, ncolor):
        tf = t[-1]
        ki = kbind

        guessed_no_ribosomes = int(1.3*ki*tf)
        result = np.zeros((len(t)*N_rib), dtype=np.int32)
        ribtimes = np.zeros((guessed_no_ribosomes), dtype=np.float64)
        frapresult = np.zeros((len(t)*N_rib), dtype=np.int32)
        coltimes = np.zeros((guessed_no_ribosomes), dtype=np.int32)
        colpointsx = np.zeros(len(kind)*(guessed_no_ribosomes), dtype=np.int32)
        colpointst = np.zeros(len(kind)*(guessed_no_ribosomes), dtype=np.float64)
        return result, ribtimes, frapresult, coltimes, colpointsx, colpointst

    @classmethod
    def __generate_vecs_lowmem(cls, k, t, N_rib, ncolor):
        tf = t[-1]
        ki = k[0]

        guessed_no_ribosomes = int(1.3*ki*tf)
        result = np.zeros((ncolor,len(t)), dtype=np.int32)
        ribtimes = np.zeros((guessed_no_ribosomes), dtype=np.float64)
        frapresult = np.zeros((len(t)*N_rib), dtype=np.int32)
        coltimes = np.zeros((guessed_no_ribosomes), dtype=np.int32)
        colpointsx = np.zeros(len(k[1:-1])*(guessed_no_ribosomes), dtype=np.int32)
        colpointst = np.zeros(len(k[1:-1])*(guessed_no_ribosomes), dtype=np.float64)
        return result, ribtimes, frapresult, coltimes, colpointsx, colpointst

    @classmethod
    def __generate_mats_lowmem(cls, ntraj, ki, t, N_rib, ncolor):
        tf = t[-1]
        guessed_no_ribosomes = int(1.3*ki*tf)
        all_results = np.zeros((ntraj, len(t),ncolor), dtype=np.int32)
        all_ribtimes = np.zeros((ntraj, guessed_no_ribosomes), dtype=np.float64)
        all_frapresults = np.zeros((ntraj, N_rib*len(t)),dtype=np.int32)
        all_collisions = np.zeros((ntraj, guessed_no_ribosomes), dtype=np.int32)
        all_nribs = np.zeros((ntraj, 1))
        all_col_points = []

        return all_results, all_nribs, all_collisions, all_frapresults, all_ribtimes, all_col_points


    @classmethod
    def __generate_mats(cls, ntraj, ki, t, N_rib, ncolor):
        tf = t[-1]
        guessed_no_ribosomes = int(1.3*ki*tf)
        all_results = np.zeros((ntraj, N_rib*len(t)), dtype=np.int32)
        all_ribtimes = np.zeros((ntraj, guessed_no_ribosomes), dtype=np.float64)
        all_frapresults = np.zeros((ntraj, N_rib*len(t)), dtype=np.int32)
        all_collisions = np.zeros((ntraj, guessed_no_ribosomes), dtype=np.int32)
        all_nribs = np.zeros((ntraj, 1))
        all_col_points = []




        return all_results, all_nribs, all_collisions, all_frapresults, all_ribtimes, all_col_points

    @classmethod
    def __generate_mats_lowmem_nostats(cls, ntraj, ki, t, N_rib, ncolor):

        all_results = np.zeros((ntraj, len(t),ncolor), dtype=np.int32)
        all_frapresults = np.zeros((ntraj, N_rib*len(t)), dtype=np.int32)
        return all_results, all_frapresults

    @classmethod
    def __generate_vecs_lowmem_nostats(cls, k, t, N_rib, ncolor):


        result = np.zeros((ncolor,len(t)), dtype=np.int32)

        frapresult = np.zeros((len(t)*N_rib), dtype=np.int32)

        return result, frapresult

    def __check_x0(self, x0):
        if len(x0) == 0:
            x0_clean = np.zeros((200), dtype=np.int32)
        else:
            if len(x0) >200:
                raise ValueError('Unrecognized initial condition, make sure the length of x0 is <= 200')

            x0_clean = np.zeros((200), dtype=np.int32)
            x0_clean[:len(x0)] = x0
        return x0_clean

    def __check_rates_trna(self, rates):
        if isinstance(rates, np.ndarray):
            pass
        else:
            rates = np.array(rates, dtype=np.int32)
        if rates.dtype !=int:
            raise ValueError('trna ID values are not integers, to run the trna ssa version, k_index needs to be an ID of 0-60 integer values')
        if len(np.where(rates < 0)[0]) or len(np.where(rates > 60)[0]):
            raise ValueError('trna ID values are out of bounds, the IDs are 0-60 for tRNA species')

    def __check_input_memview(self, inputs):
        check = False
        for item in inputs:
            try:
                item.dtype
                if item.dtype not in [np.int32]:
                    if 'int' in str(item.dtype):

                        check = True
            except:
                pass
        return check

    def __check_memview(self, inputs):
        #make sure everything is int 32, no int64
        returned_inputs = ()
        for item in inputs:
            try:
                item.dtype
                if item.dtype not in [np.int32]:
                    if 'int' in str(item.dtype):
                        item = item.astype(np.int32)
                returned_inputs = returned_inputs + (item,)

            except:
                returned_inputs = returned_inputs + (item,)
        return returned_inputs


    def __check_rates(self, rates):
        if isinstance(rates,np.ndarray):
            pass
        else:
            rates = np.array(rates)
        if len(np.where(rates < 0)[0]) > 0:
            raise ValueError('one or more model rates are negative, double check the provided rates')


    def __ssa_python(self, k, t_array,
                     inhibit_time=0, FRAP=False, Inhibitor=False,
                     flags=None, kon=1, koff=1, kprobe=1, ssa_conditions=None):
        '''
        mRNA Translation simulation python implementation

        given a propensity vector k, time array to record, 
        and inhibitory conditions, run a single trajectory
        of translation simulation

        The simulation is described here: [PUT LINK HERE TO PAPER]

        *args*

            **k**, propensity vector of size gene length + 2,
            [initiation rate,  Codon dependent rates,  completion rate / unbinding rate]
            for reference the codon dependent rates are refering to the time rate of a ribosome to move on to the next codon

            **t_array**, time points to record the ribosome posistions at

        *keyword args*

            **inhibit_time**, the time to start inhibition assays if FRAP or Inhibitor (harringtonine) == True

            **FRAP**, True or false to apply Fluorescence Recovery After Photobleaching (FRAP) https://en.wikipedia.org/wiki/Fluorescence_recovery_after_photobleaching

            **Inhibitor**, True or false to apply harringtonine at inhibit_time. Harringtonine acts as a protien translation initiation inhibitor

        '''

        #SSA params and propensities
        R = self.default_conditions['footprint'] #exclusion volume (ribosome footprint), ribosomes cant be less than 10 codons apart because of their physical size
        kelong = np.array([k[1:-1]]).T  #rates for ribosomes moving to the next codon, based on tRNA concentrations

        N = len(kelong)  #Number of codons in the mRNA
        kbind = k[0]   #rate for a ribosome to bind and start translation
        kcompl = k[-1]     #rate for a ribosome at the end of the mRNA to unbind
        X = np.array([0, 0], dtype=int)   #the updating ribosome posistion vector that is changed in the simulation

        bursting = flags[0]
        leaky = flags[1]


        if bursting:
            burst = np.random.rand() < (kon/(kon+koff) )
        else:
            burst = 1

        Ncol = np.zeros((1,0))

        N_rib = 200  #Maximum number of ribosomes on a single mRNA (hard limit for the simulation not a physical constant)


        colors = ssa_conditions['probe_loc'].shape[0]
        if leaky:
            leaky_probe_matrix = np.zeros((colors, N, N_rib))
        intensity = np.zeros((colors, len(t_array)))
        probe = np.array(np.where(ssa_conditions['probe_loc'] ==1))
        probevec = ssa_conditions['probe_vec']


        #example X arrays and how its formatted:
        # X = [423 30 10 0 ]  read from left to right theres a ribosome in position 423 30 and 10, with a 0 kept as a buffer for simulation

        t = t_array[0]  #time point
        Nt = len(t_array)  #number of time points to record over
        tf = t_array[-1]  #final time point
        col = np.zeros((1, N_rib))
        X_array = np.zeros((N_rib, Nt))  #recording array that records the ribosome posistions over time array points
        NR = 0  #number of ribosomes bound
        it = 1  #number of iterations
        Sn_p = np.eye(max(NR+1, 2), dtype=int) #stoichiometry for the SSA
        if bursting:
            wn_p = np.zeros((X.shape[0]+1, 1)) # propensities for the SSA
        else:
            wn_p = np.zeros((X.shape[0], 1)) # propensities for the SSA

        T = np.array([0, 0], dtype=float)
        ribtimes = np.array([[0,0]], dtype=float)
        col_points = []
        #wn_p = np.zeros((1,X.shape[0])).flatten()
        wshape = len(wn_p)
        Inhibit_condition = 1  #set up inhibitor flags
        while t < tf:


            if Inhibitor == True:
                if t >= inhibit_time:

                    Inhibit_condition = 0
                else:

                    Inhibit_condition = 1
            else:
                Inhibit_condition = 1
            if FRAP == True :   #if the Photobleaching is happening, "remove" ribosomes
                if t >= inhibit_time and t < inhibit_time + 20:
                    #X = np.array([0, 0])
                    a=1
                    #T = np.array([0,0])

            oldNR = NR
            NR = len(np.flatnonzero(X)) #each iteration get the number of ribosomes on the mRNA

            if X.shape[0] < NR+1:  #if the last reaction added a ribosome put a 0 on the end of X vec

                X = np.append(X, [0])
                T = np.append(T, [0])
                T[-2] = t



            X[-1] = 0
            T[-1] = 0

            X = X[0:max(NR, 1)+1]  #clear any additional 0's on the end
            T = T[0:max(NR, 1)+1]

            if oldNR != NR:     #if the number of ribosomes has changed reallocate the sizes of stoich and propensities

                if not bursting:
                    Sn_p = np.eye(max(NR+1, 2), dtype=int)
                    wn_p = np.zeros((X.shape[0], 1))
                else:
                    Sn_p = np.eye(max(NR+1, 2), dtype=int)
                    wn_p = np.zeros((X.shape[0]+1, 1))

                if leaky:
                    leaky_probe_matrix[probe[0], probe[1], NR-1] = (np.random.rand(len(probe[0])) < kprobe).astype(int)


                wshape = len(wn_p)


            Sn = Sn_p
            wn = wn_p


            #get indices of where X vecs are > 0 ie where the ribosome values are
            inds = X > 0

            if bursting:

                    wn[:-1][inds] = kelong[X[inds]-1]  #update propensities



            else:
                wn[inds] = kelong[X[inds]-1]


            if X[0] == N:  #if the ribosome in the 0 position is at the end of the mRNA set propensities to the reaction for completion

                Sn[:, 0] = (np.append(X[1:], np.array([0]))-X[0:])


                wn[0] = kcompl


            #if there are no ribosomes or when there is enough room for a new ribosome to bind add the propensity for binding
            if NR == 0:

                wn[NR] = kbind*Inhibit_condition*burst



            if NR > 0 and X[NR-1] > R:
                wn[NR] = kbind*Inhibit_condition*burst

            REST = np.less(X[1:]+10, X[0:-1])  #apply the footprint condition ie set any propensities where it violates the > 10 codons apart rule to 0

            if bursting:
                if NR > 1:
                    wn[1:-1] = (wn[1:-1].T*REST).T
                else:
                    wn[1:] = (wn[1:].T*REST).T
            else:
                wn[1:] = (wn[1:].T*REST).T  #apply that logical^ to propensities

            if bursting:
                if burst:
                    wn[-1] = koff
                else:
                    wn[-1] = kon

            w0 = sum(wn.flat)  #get the sum of propensities
            randnum = np.random.random_sample(2) #update time to point of next reaction (exponential waiting time distb)
            t = (t-np.log(randnum[0])/w0)

            while it < Nt and t > t_array[it]:  #record state if past timepoint
                X_array[0:len(X), it] = X

                if leaky:
                    int_tmp = np.zeros(colors)
                    validx = X[X>0]
                    c = leaky_probe_matrix[:, :, 0]
                    for ribind in validx:
                        int_tmp += np.sum(c[:, :(ribind-1)],axis=1)

                    intensity[:, it] = int_tmp

                else:
                    validx = X[X>0]
                    int_tmp = np.zeros(colors)
                    for ribind in validx:
                        int_tmp += probevec[:, (ribind-1)]
                    intensity[:, it] = int_tmp

                it += 1



            if t < tf:  #if still running simulation pick which reaction happened via random number and propensity sum
                r2 = w0*randnum[1]
                tmp = 0

                for i in range(wshape):

                    tmp = tmp + wn[i]
                    if tmp >= r2:
                        event = i
                        break


            if bursting:
                if event >= NR+1:
                    burst+=1
                    burst = burst%2
                else:
                    X = (X + Sn[:, event].T)

                    if np.sum(Sn[:, event]) < 0 :

                        ribtimes = np.vstack((ribtimes, [T[0],t]))
                        T[:-1] = T[1:]
                        Ncol = np.append(Ncol, col[0][0] )
                        col = np.atleast_2d(np.append(col[:,1:],[0]))

                    else:
                        if X[event-1] == X[event] + R:
                            col[0][event] +=1
                            col_points.append((X[event], t))

            else:
                X = (X + Sn[:, event].T)  #update X vector for new ribosome state


                if np.sum(Sn[:,event]) < 0 :

                    ribtimes = np.vstack((ribtimes, [T[0], t]))
                    T[:-1] = T[1:]
                    Ncol = np.append(Ncol, col[0][0])
                    col = np.atleast_2d(np.append(col[:, 1:],[0]))

                    if leaky: #remove the leaky probes
                        leaky_probe_matrix[probe[0], probe[1], 0] = leaky_probe_matrix[probe[0], probe[1], 1]


                else:
                    if X[event-1] == X[event] + R:
                        col[0][event] +=1
                        col_points.append((X[event], t))



        return X_array,ribtimes[1:, :], Ncol, col_points, intensity  #return the completed simulation




# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 17:45:31 2020

@author: willi
"""

import warnings
import numpy as np
from . import CodonDictionaries
from . import SequenceManipMethods

class PropensityFactory():
    '''
    factory class for the k's
    '''
    def __init__(self):
        self.codon_dicts = CodonDictionaries.CodonDictionaries()

    @staticmethod
    def bin_k(k, inds):
        '''
        Generate a k_binned from a given indexing vector

        Parameters
        ----------
        k : array like
            propensity list or ndarray.
        inds : ndarray
            numpy array of index locations, can be generated from even_bin()
            or intellegent_bin().

        Returns
        -------
        k_binned : list
            propensity rates for the proposed binning strategy.

        '''
        try:
            k = k.tolist()
        except:
            pass
        k_binned = np.zeros(len(inds)-1)
        binned_ks = []
        for i in range(0, len(inds)-1):
            binned_ks = binned_ks + [k[inds[i]:inds[i+1]],]


        for i in range(0, len(inds)-1):
            k_binned[i] = 1/ np.sum(1/np.array(binned_ks[i]))
        return k_binned


    def get_trna_ids(self, nt_seq):
        '''
        Return tRNA ids 0-60 for a given nt_seq

        Parameters
        ----------
        nt_seq : str
            nucleotide sequence.

        Returns
        -------
        list
            list of trna IDs 0-60 (one for each codon).

        '''

        if '*' in SequenceManipMethods.SequenceManipMethods().nt2aa(nt_seq)[-1]:
            nt_seq = nt_seq[:-3]

        codons = nt_seq.upper()
        seperated_codons = [codons[i:i+3] for i in range(0, len(codons), 3)]
        return [self.codon_dicts.trna_dict[x] for x in seperated_codons]



    def get_k(self, nt_seq, k_init, k_elong_mean,
              k_term, codon_freq_bias_dict=None):
        '''
        Returns all propensity functions for a given kinit, kelong_mu, and k termination

        Parameters
        ----------
        nt_seq : string
            nucleotide sequence.
        k_init : float
            initation rate.
        k_elong_mean : float
            average kelongation rate for the transcript.
        k_term : float
            termination rate.

        Returns
        -------
        all_k : list
            list of propensity values per location + initiation and termination at respective ends.

        '''

        if codon_freq_bias_dict is None:
            codon_dict = self.codon_dicts.human_codon_frequency_bias_nakamura
        else:
            codon_dict = codon_freq_bias_dict
        codons = nt_seq.upper()

        genelength = int(len(codons)/3)
        seperated_codons = [codons[i:i+3] for i in range(0, len(codons), 3)] #split codons by 3
        k_elongation = np.zeros((1, genelength))
        tRNA_copynumber = np.zeros((1, genelength))


        for i in range(len(seperated_codons)):
            tRNA_copynumber[0, i] = codon_dict[seperated_codons[i]]

        mean_tRNA_copynumber = np.mean(list(codon_dict.values()))

        k_elongation = (tRNA_copynumber / mean_tRNA_copynumber) * k_elong_mean
        all_k = [k_init] + k_elongation.flatten().tolist() + [k_term]

        return all_k


    def get_k_3_frame(self, nt_seq, k_elong_mean,
                      codon_freq_bias_dict=None):
        '''
        Parameters
        ----------
        nt_seq : str
            nucleotide sequence.
        k_elong_mean : float
            mean elongation rate aa/time.

        Returns
        -------
        kelongs : list
            elongation rates for all 3 frames of a nucleotide sequence stacked in:
                [0+frame L, 1+frame L-1, 2+frame L-1] .

        '''
        kelongs = []

        if codon_freq_bias_dict is None:
            codon_dict = self.codon_dicts.human_codon_frequency_bias_nakamura

        for n in range(3):
            if n != 0:
                codons = nt_seq[n:-(3-n)]
            else:
                codons = nt_seq
            genelength = int(len(codons)/3)
            seperated_codons = [codons[i:i+3] for i in range(0, len(codons), 3)] #split codons by 3
            k_elongation = np.zeros((1, genelength))
            tRNA_copynumber = np.zeros((1, genelength))


            for i in range(len(seperated_codons)):
                tRNA_copynumber[0, i] = codon_dict[seperated_codons[i]]

            mean_tRNA_copynumber = np.mean(list(codon_dict.values()))

            k_elongation = (tRNA_copynumber / mean_tRNA_copynumber) * k_elong_mean

           # k_elongation.flatten().tolist()[:-1]

            kelongs = kelongs + k_elongation.flatten().tolist()

        return kelongs


    @staticmethod
    def get_binned_k(k, bins):
        '''
        given a propensity vector and an desired binning indices, return pv and pl binned

        Parameters
        ----------
        k : 1 x L  numpy array
            Propensity vector of rates.
        bins : 1xL ind locations sizes
            binning strategy, the sum of this vector should = L.

        Returns
        -------
        propensity_binned: 1 x Nbins numpy array
            the binned propensities.
        bin_sizes: 1 x Nbins numpy array
            list of bin sizes.

        '''
        binsize = int(np.floor(len(k)/bins))
        binned_ks = []

        k_binned = np.zeros(bins)
        k_lens = np.ones(bins)*binsize

        to_redistribute = len(k)%bins

        k_lens[-to_redistribute:] = binsize+1

        inds = np.hstack(([0.], np.cumsum(k_lens))).astype(int)


        for i in range(0, bins):
            binned_ks = binned_ks + [k[inds[i]:inds[i+1]].tolist(),]

        for i in range(0, bins):
            k_binned[i] = np.mean(binned_ks[i])/len(binned_ks[i])

        return k_binned, k_lens


    @staticmethod
    def intellegent_bin(pl, nbins, min_bin=1):
        '''
        Function to do intellegent binning, focuses resolution on the areas
        defined in the probe location vector

        Note if you pass it a minium bin that when min_bin*nbins > length of your sequence
        this function will fail

        Parameters
        ----------
        pl : numpy array
            numpy array of Ncolor x Length 0's 1's for probe locations.
        nbins : int
            number of desired bins.
        min_bin : int, optional
            DESCRIPTION. The default is 1.

        Returns
        -------
        pl_inds : numpy array
            n bin locations over the vector length given.

        '''

        if min_bin*nbins > pl.shape[1]:
            warnings.warn('Desired minimum binsize and desired number'\
                          ' of bins is not possible with the length of'\
                              ' the probe vector, returning best guess')

        pl_inds = np.where(pl == 1)[1]

        if 0 not in pl_inds:
            pl_inds = np.hstack((np.array([0]), pl_inds))
        if pl.shape[1]+1 not in pl_inds:
            pl_inds = np.hstack((pl_inds, np.array(pl.shape[1]+1)))

        used_bins = len(pl_inds)
        k = len(pl_inds)-1
        j = 0
        to_add = []
        while used_bins < nbins+1:
            if j == k:

                j = 0

                prev_pl_inds = pl_inds
                pl_inds = np.hstack((pl_inds, np.array(to_add)))
                pl_inds = np.sort(pl_inds)

                if np.array_equal(prev_pl_inds, pl_inds):
                    break
                k = len(pl_inds)-1
                to_add = []

            newbin = int(pl_inds[j] + (pl_inds[j+1]-pl_inds[j])/2)

            if newbin not in pl_inds:
                if not (np.abs(pl_inds - newbin) <= min_bin).any():
                    to_add.append(newbin)
                    used_bins += 1
            j += 1

        pl_inds = np.hstack((pl_inds, np.array(to_add)))
        pl_inds = np.sort(pl_inds)

        return pl_inds

    @staticmethod
    def even_bin(length, nbins):
        '''
        Parameters
        ----------
        length : int
            Length of the vector to bin.
        nbins : int
            Number of desired bins.

        Returns
        -------
        inds : numpy array
            n bin locations over the vector length given.


        '''
        binsize = int(np.floor(length/nbins))

        k_lens = np.ones(nbins)*binsize

        to_redistribute = length%nbins

        k_lens[-to_redistribute:] = binsize+1

        inds = np.hstack(([0.], np.cumsum(k_lens))).astype(int)

        return inds
    


# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 17:52:28 2020

@author: Micheal May, Zach Fox and Will Raymond
"""
import numpy as np
from scipy.sparse.linalg import expm
from scipy.integrate import odeint
#from scipy.integrate import ode
import scipy as sp
import scipy.optimize
#import random as random
import collections as c
#import matplotlib.pyplot as plt
import copy as cp
#from multiprocessing import process as Process
import matplotlib.pyplot as plt

'''
Generic solvers written by Micheal May, Zach Fox and Will Raymond ~2018
'''


class GenericSSA:
    def __init__(self,type='linear'):
        self.xi=np.array([])
        self.ti=None
        self.tf=None
        self.S=np.array([])
        self.type=type
        self.ptimes=100
        self.params={}
        if type=='linear':
            self.W0=np.array([])   
            self.W1=np.array([])
        if type == 'nonlinear':
            self.P=lambda x,t:None 
    def gettvec(self):
        return np.linspace(self.ti,self.tf,self.ptimes)
        
    def updateparams(self):
        for k, v in self.params.iteritems():
            setattr(self, k, v)

    def _run_trajectory(self):#!!!!!!!!!!!!!!!!!renamed run to solve(big deal)
        x=self.xi
        t=0
        __n=len(x)
        self.time=self.gettvec()
        data=np.zeros((len(self.xi),self.ptimes))
        ip=0
        if self.type=='linear':
            while t<self.tf:
                rate=np.atleast_2d(np.dot(self.W1,x))+self.W0
                rate=np.cumsum(rate)
                t=(t-np.log(np.random.rand(1))/rate[-1])
                ro=rate[-1]*np.random.rand()
                while t>self.time[ip]:
                    if t>self.tf:
                        b = len(self.time[ip:])
                        fill = np.repeat(x,b)
                        data[:,ip:]=fill.reshape(__n,b)
                        return data
                    else:
                        data[:,ip]=x.reshape(__n)
                        ip=ip+1
                for i in range(len(rate)):
                    if rate[i]>=ro:
                        event=i
                        break
                x=x+np.atleast_2d(self.S[:,event]).T
        elif self.type=='nonlinear':
            pass
        else:
            'Error'
        return data

    def _solve(self,n):
        soln=np.zeros((len(self.xi),self.ptimes,n))
        for i in range(n):
            d=self._run_trajectory()
            soln[:,:,i]=d
        self.soln = soln
        return soln
        
    def getW0(self):
        pass
    
    def getW1(self):
        pass
    
    def setpar(self,key,val):
        self.params[key]=val

class GenericFSP:
    def __init__(self,ti=[],tf = [],xi=[],A=np.array([]),ptimes=100):
        self.ti=ti
        self.tf=tf
        self.xi=xi
        self.A=A
        self.errors=np.empty(ptimes)
        self.ptimes=ptimes
        self.params={}
        
    def updateparams(self):
        for k, v in self.params.iteritems():
            setattr(self, k, v)
        
    def gettvec(self):
        '''
        get vector of times 
        '''
        return np.linspace(0,self.tf-self.ti,self.ptimes)

    def _solve(self):
        '''
        solve the FSP at each time. 
        '''
        tvec=self.gettvec()
        self.model_solns = np.zeros((len(self.xi),len(tvec)))
        for i in range(len(tvec)):
            mexp = expm(self.A*tvec[i])
            self.soln = mexp.dot(self.pi)
#           self.errors[i]=1.0-np.sum(self.model_solns[:,i])
        return self.soln

class GenericODE:
    def __init__(self,ti=None,tf=None,ODE=None,ptimes=50,xi=None):
        self.ti=ti
        self.tf=tf
        self.ODE=ODE
        self.ptimes=ptimes
        self.xi=xi
        self.tvec=None
        self.soln=None
        self.parameters={}
    
    def updateparams(self):
        try:
            for k, v in self.params.items():
                setattr(self, k, v)
        except:
             for k, v in self.params.iteritems():
                setattr(self, k, v)           
            
    def gettvec(self):
        '''
        get vector of times
        '''
        return np.linspace(self.ti,self.tf,self.ptimes)

    def _solve(self):
        '''
        solve the ODEs at each time
        '''
        self.tvec=self.gettvec()
        try: 
            solution=odeint(self.ODE,self.xi,self.tvec,Dfun=self.Jacobian)
        except: 
            solution=odeint(self.ODE,self.xi,self.tvec)
        self.soln = solution.T
        return solution.T
        
    def _expm_solve(self):
        self.tvec = self.gettvec()
        
        
        
class GenericStats():
    '''
    a class of static functions to analyze data
    '''
    def __init__(self,model=None):
        if model is not None:
            self.model=model
            self.type=model.__class__.__name__
            self.simdata={}
            self.simdata['data']=self.model.solve()
            self.simdata['time']=self.model.gettvec()
            
            self.means=self.get_means(**self.simdata)
            self.acc=self.get_acc(self.simdata)
            if self.__class__.__bases__ is 'GenericSSA' or self.__class__.__name__ is 'GenericSSA':
                self.var=self.get_var(self.simdata['data'])

    @staticmethod
    def get_var(data,time):
        '''
        a function that returns statistical variance of data
        '''
        if type(data) is not np.ndarray :
            print("Error: data is not a numpy array")
            return
        s=data.shape
        if len(s) is 1:
            s=s+1
        var=np.zeros((s[0],s[1]))
        for j in range(len(time)):
            for i in range(s[0]):
                var[i,j]=np.var(data[i,j,:])
        return var
        
    @staticmethod
    def get_means(data,time):
        '''
        a function that returns the statistical mean of data
        '''
        if type(data) is not np.ndarray :
            print("Error: data is not a numpy array")
            return
        __s=data.shape
        means=np.zeros((__s[0],__s[1]))
        for j in range(len(time)):
            for i in range(__s[0]):
                means[i,j]=np.mean(data[i,j,:])
        return means
    
    @staticmethod 
    def get_acc2(data,trunc=True): 
        N = len(data)    
        fvi = np.fft.fft(data,n=2*N)     
        acf = fvi*np.conjugate(fvi)
        acf = np.fft.ifft(acf)
        acf = np.real(acf[:N])/float(N)
        if trunc:
            acf[acf<0]=0
            for i in range(1,len(acf)): 
                if acf[i] > acf[i-1]:
                    acf[i] = acf[i-1]
        return acf
        
    @staticmethod 
    def get_fastacc2(data,trunc=True): 
        N = len(data)    
        fvi = np.fft.fft(data,n=2*N)     
        acf = fvi*np.conjugate(fvi)
        acf = np.fft.ifft(acf)
        acf = np.real(acf[:N])/float(N)
        if trunc:
            acf[acf<0]=0
            for i in range(1,len(acf)): 
                if acf[i] > acf[i-1]:
                    acf[i] = acf[i-1]
        return acf
    @staticmethod
    def get_acc(data,one_sided=True,trunc=True):
        '''
        a function that return the statistical autocorreletion of data
        '''
        half = len(data)/2
        z = np.correlate(data[:half+1],data,mode='valid') / float(half)
        if trunc:
            z[z<0]=0
            for i in range(1,len(z)): 
                if z[i] > z[i-1]:
                    z[i] = z[i-1]
        return z 

    @staticmethod
    def get_xcc(data1,data2=None,axis=1,norm=False,trunc=False):
        '''
        a function that returns the statistical crosscorrelation(or autocorrelation)of data
        '''
        if data2 is None:
            data2=data1
        if type(data1) is not np.ndarray:
            return "Type error, data must be a numpy array"
        corr = np.correlate(data1,data2,mode='full')
        if norm and corr.max()>0:
            corr = corr/corr.max()
        lc = len(corr)/2
        if trunc:
            corr[corr<0]=0
            tmp = corr[lc:]
            for i in range(1,lc): 
                if tmp[i] > tmp[i-1]:
                    tmp[i] = tmp[i-1]
            corr[lc:] = tmp
        return corr 

    @staticmethod
    def get_kld(p,q):
        '''
        returns the kullback-leibler divergence between the distributions 
        p and q. 
        '''
        support = (p>0)*(q>0)
        #if np.max(self.target)<np.min(self.predicted) or np.min(self.target)>np.max(self.predicted):
        #    kld = np.inf
        kld = np.sum( p[support] * (np.log(p[support]) - np.log(q[support])))
        return kld    
    
    @staticmethod
    def normalize(data):
        return data/sum(data)
        
    @staticmethod
    def get_curvature(y,h):
        k=np.zeros(len(y)-2)
        for i in range(len(y)-2):
            k[i]=((y[i+2]-2*y[i+1]+y[i])/(h**2.0))/(1+((y[i+2]-y[i])/(2*h))**2)**(1.5)
        return k
        
    @staticmethod
    def get_sse(data1,data2,axis=None):
        '''calculates sum of squared error of two numpy array datasets'''
        return np.linalg.norm(data1-data2)
        
class GenericOpt:
    def __init__(self,model):
        self.model=model
        self.modeltype=model.__class__.__name__
        self.activeobjfun='chisq'
        self.activemethod='MH'
        self.activefun='solve'
        self._mdt={'met_hast':self.methast,\
                      'methast':self.methast,\
                      'MH':self.methast,\
                      'lin_opt':self.lin_opt,\
                      'linopt':self.lin_opt,\
                      'genetic':self.genetic,\
                      'sim_anneal':self.sim_anneal,\
                      'simanneal':self.sim_anneal}
        self._odt={'sse':self.sse,\
                   'SSE':self.sse,\
                   'chi':self.chisq,\
                   'chisq':self.chisq}
        self.opts={'bounds':([0,10],)*len(self.model.params)}
        self.data=[]
        self.time=[]
        self.cache=c.deque([])
        self.bestpars=[]
    def fit(self,*args):
        #if method is not None:
        #    return self.methods[method](self,args)
        if self.activemethod is not None and self.activeobjfun is not None:
            print("Running: " +self.activemethod +" with "+self.activeobjfun)
            return self._mdt[self.activemethod](*args)
        else:
            print("Please pick an active solver and objective function")
            
    def getsimdata(self):
        return self.model.solve()
    
    def sse(self):
        simdata=self.getsimdata()
        return np.sum((self.data-simdata)**2)

    def chisq(self):
        simdata=self.getsimdata()
        return np.sum(((self.data-simdata)**2)/self.model.variance)
    
    def kld(self,x):
        pass#return np.stats.kld(1,2,3)
    def methast(self,N=1000,burn=100,stepsize=5):
        def evolvepars(p,stepsize):
            new=cp.copy(p)
            for j in range(len(new)):
                while 1 :
                    new[j]=p[j]+np.random.normal(0,stepsize)
                    #print p
                    #print new
                    if (self.opts['bounds'][j][0]<new[j]<self.opts['bounds'][j][1]):
                        break
            return new
                        
        optfun=self.getoptfun()
        oldpars=self.bestpars=self.getpars()
        f_old=f_best=optfun(oldpars)   
        for i in range(-burn,N):
            newpars=evolvepars(oldpars,stepsize)
            f_new=optfun(newpars)
            #print("newpars = "+str(newpars)+ "     oldpars = "+str(oldpars))
            #print("fnew="+str(f_new)+", fold="+str(f_old))
            if f_new<f_old or np.random.rand()<np.exp(f_old-f_new):
                #raw_input("Press Enter to continue...")
                f_old=f_new
                oldpars=newpars
                print("Accpeted "+str(i))
                if i>0:
                    self.cacheitem(newpars)    
                if f_new<f_best:
                    f_best=f_new
                    self.bestpars=newpars
                

        return self.bestpars
    
    def genetic(self):
        f=self.getoptfun()
        logfun = lambda x: f(x)
        return sp.optimize.differential_evolution(logfun,**self.opts)
    def lin_opt(self):
        pass
    def sim_anneal(self):
        return sp.optimize.minimize(self.getoptfun(),self.getpars(),method='Anneal',options={'lower':[0,0]})
    
    def getoptfun(self):
        def fun(pars):
            self.setpars(pars)
            #print self.model.params
            return self._odt[self.activeobjfun]()
        return fun
        
    def getpars(self):
        i=0
        pars=np.zeros(len(self.model.params))
        for key,val in self.model.params.iteritems():
                pars[i]=val
                i+=1
        return pars
    def clearcache(self):
        self.cache=c.deque([])
    def cachepars(self):
        self.cache.append((self.getpars()))
    def cacheitem(self,item):
        self.cache.append(item)
    def setpars(self,parval):
        i=0
        for key,val in self.model.params.iteritems():
                self.model.params[key]=parval[i]
                i+=1
    def parplot(self,s=2,alpha=.5):
        N=len(self.model.params)
        D=zip(*self.cache)
        #print D

        for i in range(N):
            for j in range(N):
                plt.subplot(N,N,j+N*i+1)
                if i is not j:                    
                    plt.scatter(D[i],D[j],linewidth=0.0,s=s,alpha=alpha)
                else:
                    plt.hist(D[i])
        
class GenericModel(GenericSSA,GenericODE):
    '''
    define a generic model to run SSA,ODE~may add FSP analysis here one day
    '''
    def __init__(self,ti=None,tf=None,ODE=None,ptimes=50,xi=None):
        self.ti=ti
        self.tf=tf
        self.ODE=ODE
        self.ptimes=ptimes
        self.xi=xi
        self.tvec=None
        self.soln=None
        self.parameters={}
        self.type=None
    def updateparams(self):
        for k, v in self.params.iteritems():
            setattr(self, k, v)
            
    def gettvec(self):
        '''
        get vector of times
        '''
        return np.linspace(self.ti,self.tf,self.ptimes)

    def _solve(self):
        '''
        solve the ODEs at each time
        '''
        self.tvec=self.gettvec()
        try: 
            solution=odeint(self.ODE,self.xi,self.tvec,Dfun=self.Jacobian)
        except: 
            solution=odeint(self.ODE,self.xi,self.tvec)
        self.soln = solution.T
        return solution.T
    



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 18:08:58 2020

@author: willi
"""

import numpy as np
from . import PropensityFactory, ProbeVectorFactory
PropensityFactory = PropensityFactory.PropensityFactory
ProbeVectorFactory = ProbeVectorFactory.ProbeVectorFactory
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib import cm
import matplotlib.patches as mpatches
import time


path_to_cpp = ''
path_to_gen = ''
path_to_trna = ''

import os
for root, dirs, files in os.walk(".", topdown=False):
   for branch in dirs:
       if 'ssa_cpp' in branch:
           path_to_cpp = os.path.join(root, branch)
       if 'generalized_cpp' in branch:
           path_to_gen = os.path.join(root, branch)
       if 'trna_ssa' in branch:
           path_to_trna = os.path.join(root, branch)
if path_to_cpp != '':
    try:
        cwd = os.getcwd()
        os.chdir(path_to_cpp)
        
        import ssa_translation
        import ssa_translation_lowmem
        os.chdir(cwd)
    except:
        os.chdir(cwd)
    

if path_to_gen != '':
    try:
        cwd = os.getcwd()
        
        os.chdir(path_to_gen)
        print('importing C++ models')
        import ssa_translation_generic
        import ssa_translation_generic_lowmem
        print('c++ models loaded successfully')
        os.chdir(cwd)
    except:
        os.chdir(cwd)

if path_to_trna !='':
    try:
        cwd = os.getcwd()
        
        os.chdir(path_to_gen)
        print('importing C++ tRNA models')
        import ssa_trna
        print('c++ models loaded successfully')
        os.chdir(cwd)
    except:
        os.chdir(cwd)   
        
        
    
try: 
    import ssa_translation_lowmem
except:
    pass

class ModelBuilder():
    '''
    Container class for the solvers
    '''
    def __init__(self, time=None,xi=None):
        self.k = None
        self.k_bind = None
        self.k_term = None
        self.multiframe = True
        self.additional_rxns = {}
        self.probe_locations = None
        self.colors = 1
        
        self.default_conditions = {'low_mem':True,
                                   'perturb':[0,0,0],
                                   'leaky_probes':False,
                                   'bins':None,
                                   'k_probe':0,
                                   'footprint':9,
                                   'burnin':500,
                                   'record_stats':False}
        
        self._poi = None
        self.k_enters = np.array([])
        self.k_stops = np.array([])
        self.k_jumps = np.array([])
        self.k_pauses = np.array([])
        self.frames_used = 1
        self.probe = None
        
        
    @property    
    def n_extra_rxns(self):
        return len(self.k_enters) + len(self.k_stops) + len(self.k_jumps) + len(self.k_pauses) 
        
    def add_k(self, poi, kelong_mean = 10):
        self._poi = poi
        ks = PropensityFactory().get_k_3_frame(poi.nt_seq, kelong_mean)        
        self.k = ks

        L = len(poi.aa_seq)
        self.k[L-1] = 0
        self.k[L*2-1  -1] = 0
        self.k[L*3-2 -1] = 0
        self.k_stops = np.array([[L,0,10],[L-1,1,10],[L-1,2,10]],dtype=np.float64)
        
    
    def add_jumps(self, locs, frames, dest_locs, dest_frames, rates):
        if isinstance(locs,int):
            locs = [locs]
        if isinstance(frames,int):
            frames = [frames]
        if isinstance(dest_locs,int):
            dest_locs = [dest_locs]
        if isinstance(dest_frames,int):
            dest_frames = [dest_frames]
        if isinstance(rates,float):
            rates = [rates]
        self.__check_frames(frames)
        self.k_jumps = np.array((locs,frames,dest_locs,dest_frames,rates )).T.astype(np.float64)
        
    
    def add_enters(self,locs, frames, rates):
        if isinstance(locs,int):
            locs = [locs]
        if isinstance(frames,int):
            frames = [frames]
        if isinstance(rates,float):
            rates = [rates]        
        self.__check_frames(frames)
        self.k_enters = np.array((locs,frames,rates )).T.astype(np.float64)
    
    def add_stops(self,locs, frames, rates):
        if isinstance(locs,int):
            locs = [locs]
        if isinstance(frames,int):
            frames = [frames]
        if not isinstance(rates,list):
            rates = [rates]        
        self.__check_frames(frames)

        self.k_stops = np.array((locs,frames,rates )).T.astype(np.float64)
        
    
    def clear_model(self):
        self._poi = None
        self.k_enters = np.array([])
        self.k_stops = np.array([])
        self.k_jumps = np.array([])
        self.k_pauses = np.array([])
        self.frames_used = 1        
        self.probe = None
    
    def add_pauses(self,locs, frames, rates):
        if isinstance(locs,int):
            locs = [locs]
        if isinstance(frames,int):
            frames = [frames]
        if isinstance(rates,float):
            rates = [rates]        
            
        self.__check_frames(frames)
        self.k_pauses = np.array((locs,frames,rates )).T.astype(np.float64)
    
    def __check_frames(self,frames):
        for frame in frames:
            if frame > self.frames_used:
                self.frames_used = frame
        
    def visualize_transcript(self):
        
        probe = self._poi.probe_loc
        #self._poi.generate_3frame_tags()
        
        keys = []
        for dics in self._poi.multiframe_epitopes:
            for key in dics.keys():
                if key not in keys:
                    keys.append(key)
        
        
        fig,ax = plt.subplots(1,1,dpi=300)
        N = len(self._poi.aa_seq)
        ncolors = len(keys)
        xs = [.1, 2.1, 4.1][::-1]
        fr = ["+0","+1","+2"]
        
        
        opt = {'head_width': .1, 'head_length': .1,
        'length_includes_head': True}
        
        
        cmap = cm.get_cmap('viridis')
        colors = cmap(np.linspace(.01,.95, ncolors))
        
        go_color = '#2A9D8F'
        stop_color = '#E76F51'
        jump_color = '#843b62'
        pause_color = '#e9c46a'
        
        
        
        if ncolors <4:
            colors = ['#64ff21','#08ffff','#fb00ff']
        
        for i in range(3):
            

            
            rectangle =  mpatches.Rectangle((0,xs[i]), N ,.7,linewidth=1,edgecolor='k',facecolor='darkgray')
    
            ax.add_patch(rectangle)
            
            color = []
            color_dict = self._poi.multiframe_epitopes[i]
            location = []
            for n in range(len(keys)):
                if keys[n] in color_dict.keys():
                
                    location.append(color_dict[keys[n]])
                    color.append( [n for x in color_dict[keys[n]]])
                
          
            
            colorlabels = ['Color %d'% j for j in range(ncolors)    ]
            
        
            color = [j for i in color for j in i]
            location = [j for i in location for j in i]
            
            for c in range(ncolors):
                ax.plot([-10,-10],[-4,-4],color = colors[c]  )  #fix the legend colors
            
            
            for c,loc in zip(color,location):
                ax.plot([loc,loc],[xs[i] ,xs[i]+.8],color = 'k',lw=2 )
                ax.plot([loc,loc],[xs[i] ,xs[i]+.8],color = colors[c]  )
                
            ax.set_ylim([-.1,8])
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
            ax.spines['bottom'].set_visible(False)
            ax.spines['left'].set_visible(False)
            ax.set_xlabel('codon')
            
            ax.axes.get_yaxis().set_visible(False)
            if i == 0:
                leg1 = ax.axes.legend(colorlabels,loc=9)
                ax.text(-20,7.5,'Transcript Name: %s' % self._poi.name)
                ax.text(-20,6.7,'Total Length: %d aa' % self._poi.total_length)
                ax.text(-20,5.9,'Seq: %s ...' % self._poi.aa_seq[:10])
                

                
            ax.text(-N*.07 ,xs[i] + .4,fr[i])
                
        for j in range(self.k_enters.shape[0]):
            yloc = xs[int(self.k_enters[j][1])]
            loc = int(self.k_enters[j][0])
            ax.plot([loc,loc],[yloc, yloc+.8],color = 'k',lw=3  )
            ax.plot([loc,loc],[yloc, yloc+.8],color = go_color,lw=2  )

            ax.annotate("",
                    xy=(loc, yloc+.85), xycoords='data',
                    xytext=(loc-1, yloc+.85), textcoords='data',
                    arrowprops=dict(arrowstyle="-|>",lw=1.5,facecolor='k',edgecolor='k',
                                    connectionstyle="arc3"),
                    )            
            
            ax.annotate("",
                    xy=(loc, yloc+.85), xycoords='data',
                    xytext=(loc-1, yloc+.85), textcoords='data',
                    arrowprops=dict(arrowstyle="-|>",lw=1,facecolor=go_color,edgecolor=go_color,
                                    connectionstyle="arc3"),
                    )
            

        for j in range(self.k_stops.shape[0]):
            yloc = xs[int(self.k_stops[j][1])]
            loc = int(self.k_stops[j][0])
            ax.plot([loc,loc],[yloc, yloc+.8],color = 'k',lw=3  )
            ax.plot([loc,loc],[yloc, yloc+.8],color = stop_color,lw=2  )
            
        for j in range(self.k_jumps.shape[0]):
            
            yloc1 = xs[int(self.k_jumps[j][1])]
            loc1 = int(self.k_jumps[j][0])
            yloc2 = xs[int(self.k_jumps[j][3])]
            loc2 = int(self.k_jumps[j][2])   

            ax.plot([loc1,loc1],[yloc1, yloc1+.8],color = 'k',lw=3  )
            ax.plot([loc2,loc2],[yloc2, yloc2+.8],color = 'k',lw=3  )
            
            ax.plot([loc1,loc1],[yloc1, yloc1+.8],color = jump_color,lw=2  )
            ax.plot([loc2,loc2],[yloc2, yloc2+.8],color = jump_color,lw=2  )
            
            if yloc2 == yloc1:
                ax.plot([loc1,loc1],[yloc1, yloc1-.3],color=plt.rcParams['axes.edgecolor'])
                ax.plot([loc1,loc2],[yloc1-.3, yloc1-.3],color=plt.rcParams['axes.edgecolor'])
                
                ax.annotate("",
                        xy=(loc2, yloc2), xycoords='data',
                        xytext=(loc2, yloc1-.3), textcoords='data',
                        arrowprops=dict(arrowstyle="-|>",lw=2,
                                        connectionstyle="arc3"),
                        )
                
            else:
                dx = loc2 - loc1
            
                dy = yloc2- yloc1 
                
                if yloc1 > yloc2:
                    dx = loc2 - loc1
            
                    dy = yloc2+.8- yloc1 
                
                   # ax.arrow(loc1,yloc1,dx,dy,color = 'k',lw=2)#,  head_width=5, head_length=.2)
                    ax.annotate("",
                        xy=(loc2, yloc2+.7), xycoords='data',
                        xytext=(loc1, yloc1), textcoords='data',
                        arrowprops=dict(arrowstyle="-|>",lw=2,
                                        connectionstyle="arc3"),
                        )
                else:
                    dx = loc2 - loc1
            
                    dy = yloc2- yloc1+.8
                    #ax.arrow(loc1,yloc1,dx,dy,color = 'k',lw=1)
                    
                    ax.annotate("",
                        xy=(loc2, yloc2), xycoords='data',
                        xytext=(loc1, yloc1+.8), textcoords='data',
                        
                        
                        arrowprops=dict(arrowstyle="-|>",lw=2,
                                        connectionstyle="arc3"),
                        )

        for j in range(self.k_pauses.shape[0]):
            yloc = xs[int(self.k_pauses[j][1])]
            loc = int(self.k_pauses[j][0])
            ax.plot([loc,loc],[yloc, yloc+.8],color = 'k',lw=3  )
            ax.plot([loc,loc],[yloc, yloc+.8],color = pause_color,lw=2  )         
            
            
            
            
        custom_lines = [Line2D([0], [0], color=go_color, lw=2),
                Line2D([0], [0], color=stop_color, lw=2),
                Line2D([0], [0], color=pause_color, lw=2), 
                Line2D([0], [0], color=jump_color, lw=2)]
        
        ax.legend(custom_lines, ['Enter', 'Stop', 'Pause','Jump'])
        ax.add_artist(leg1)
        #ax.set_facecolor('darkgray')
        fig.show()    
        
        

    def __generate_additional_ks(self,enters,pauses,jumps,stops,L):
        
        def frame_check_1(L,arr):        
            return (L- arr[:,1]+1)*(arr[:,1]>0) + L*(arr[:,1]>1)
        
        def frame_check_3(L,arr):        
            return (L- arr[:,3]+1)*(arr[:,3]>0) + L*(arr[:,3]>1)            
                    
        def gen_ks_1_loc(L,arr):
            arr[:,0] = arr[:,0]+frame_check_1(L,arr)
            arr[:,1] = arr[:,2]    
            arr = arr[:,0:2]
            max_arr = np.max( arr[:,0])     
            return arr,max_arr
        
        def gen_ks_3_loc(L,arr):
            arr[:,0] = arr[:,0]+ frame_check_1(L,arr)     
            arr[:,1] = arr[:,2]+ frame_check_3(L,arr)
            arr[:,2] = arr[:,4]
            arr = arr[:,0:3]
            max_arr = max([np.max( arr[:,0]),np.max( arr[:,1])])
            return arr,max_arr
    
        max_enter = 0
        max_pause = 0
        max_stop = 0
        max_jump = 0
        k_jumps = np.copy(jumps)
        k_pauses = np.copy(pauses)
        k_stops = np.copy(stops)
        k_enters = np.copy(enters)
        if len(k_enters) != 0:
            k_enters,max_enter = gen_ks_1_loc(L,k_enters)
    
        if len(k_pauses) != 0:
            k_pauses,max_pause = gen_ks_1_loc(L,k_pauses)
    
        if len(k_stops) != 0:
            k_stops,max_stop = gen_ks_1_loc(L,k_stops)
        
        if len(k_jumps) != 0:
            k_jumps,max_jump = gen_ks_3_loc(L,k_jumps)
            
        max_loc = max(max_jump,max_stop,max_pause,max_enter)
        
        if max_loc <=L: 
            frames_used = 0
        if max_loc > L:
            frames_used = 1
        if max_loc > 2*L-1 :
            frames_used = 2
        
        return k_enters, k_pauses, k_stops, k_jumps, frames_used            

    def generate_probe(self):
        keys = []
        for frame in self._poi.multiframe_epitopes:
            for key in frame.keys():
                if key not in keys:
                    keys.append(key)
                    
        ncolors = len(keys)
        
        N = len(''.join(self._poi.multiframe_aa_seq))
        Ns = [0,len(self._poi.multiframe_aa_seq[0]),len(self._poi.multiframe_aa_seq[1])+len(self._poi.multiframe_aa_seq[0])   ]
        probe = np.zeros((ncolors,N) )
        
        for j in range(3):
            dic = self._poi.multiframe_epitopes[j]
            for i in range(len(keys)):
                if keys[i] in dic.keys():
                    epis = dic[keys[i]]
                    
                    probe[i,Ns[j] + np.array(epis)] = 1
        return probe

    def run_ssa(self,t, n_traj=10,low_mem = True ):
        
        try: 
            self.probe.shape
        except:
            self.probe = self.generate_probe()
        
        
        probe = self.probe.astype(int).copy(order='C')
        t_array = t
        t0 = t[0]
        ncolors = probe.shape[0]

        N_rib = 200
        #result = np.zeros((len(t_array)*N_rib),dtype=np.int32  )
        #kelong = np.array([3.1,3.2,3.3,3.4,3.5,3.1,3.2,3.3,3.4,3.5],dtype=np.float64)
        n_trajectories = n_traj
        start = time.time()
        
        lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))
        
        all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)
        
        if low_mem == False:
            all_results = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)
            all_ribtimes = np.zeros((n_trajectories,400),dtype=np.float64)
            all_coltimes = np.zeros((n_trajectories,400),dtype=np.int32)
            nribs = np.array([0],dtype=np.int32)
            all_ribs = np.zeros((n_trajectories,1))
            
        else:
            all_results = np.zeros((n_trajectories,ncolors,len(t_array)),dtype=np.int32)
        
        seeds = np.random.randint(0,0x7FFFFFF,n_trajectories, dtype= np.int32)
        
        k_enters,k_pauses,k_stops,k_jumps,frames_used = self.__generate_additional_ks(self.k_enters,self.k_pauses,self.k_jumps,self.k_stops,self._poi.total_length)
        
        #k_add = np.hstack((self.k_enters.flatten(),self.k_pauses.flatten(),self.k_stops.flatten(),self.k_jumps.flatten() ))
        k_add = np.hstack((k_enters.flatten(),k_pauses.flatten(),k_stops.flatten(),k_jumps.flatten() ))
        
        

        max_ribs = 200
        
        

        n_enters = self.k_enters.shape[0]
        n_pauses = self.k_pauses.shape[0]
        n_stops = self.k_stops.shape[0]
        n_jumps = self.k_jumps.shape[0]
        

                
        if low_mem == False:
            kelong = np.array(self.k).astype(np.float64)
            for i in range(n_trajectories):
                result = np.zeros((len(t_array)*N_rib),dtype=np.int32)    
                frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
                
                ribtimes = np.zeros((400),dtype=np.float64)
                coltimes = np.zeros((400),dtype=np.int32)
                ssa_translation_lowmem.run_SSA_generic(result,ribtimes,coltimes, kelong,frapresult,t_array, np.array([0,0,0],dtype=np.float64), seeds[i],nribs, k_add.flatten()  ,n_enters,n_pauses,n_stops,n_jumps)
                all_results[i,:] = result
                all_frapresults[i,:] = frapresult
                all_coltimes[i,:] = coltimes
                all_ribtimes[i,:] = ribtimes
                all_ribs[i,:] = nribs[0]
            
        else:
            
            
        
            kelong = np.array(self.k).astype(np.float64)
            
            for i in range(n_trajectories):
                result = np.zeros((ncolors,len(t_array)),dtype=np.int32)    
                frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
                
                

                ssa_translation_lowmem.run_SSA_generic_lowmem(result,kelong ,
                                                              frapresult,
                                                              t_array, 
                                                              np.array([0,0,0],dtype=np.float64),
                                                              int(seeds[i]),
                                                              k_add.flatten()  ,
                                                              int(n_enters),
                                                              int(n_pauses),
                                                              int(n_stops),
                                                              int(n_jumps), 
                                                              probe.astype(np.int32),
                                                              int(ncolors) ,
                                                              int(max_ribs) )
                all_results[i,:,:] = result
                #all_frapresults[i,:] = frapresult
        

      
        #traj = all_results[0,:].reshape((N_rib,len(t_array))).T
        return all_results
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Thu Nov 29 19:42:52 2018

@author: willi
"""

from . import CodonDictionaries
cdict = CodonDictionaries.CodonDictionaries()

from . import AuxCodonDicts
aux_cdict = AuxCodonDicts.AuxCodonDicts()

from . import DiffusionRateCalc
diffcalc = DiffusionRateCalc.DiffusionRateCalc()


from . import SequenceManipMethods 
seqmanip = SequenceManipMethods.SequenceManipMethods()

from . import IntensityAnalyses
inta = IntensityAnalyses.IntensityAnalyses()

from . import IntensityAnalysesRagged
inta_r = IntensityAnalysesRagged.IntensityAnalysesRagged()

from . import expv
expv = expv.expv

from . import poi
poi = poi.poi

from . import FileParser
fp = FileParser.FileParser()

from . import ProbeVectorFactory
probef = ProbeVectorFactory.ProbeVectorFactory()

from . import PropensityFactory
propf = PropensityFactory.PropensityFactory()

from . import ModelBuilder
model_builder = ModelBuilder.ModelBuilder()

from . import TranslationSolvers 
solver = TranslationSolvers.TranslationSolvers()

from . import tRNAmap
trna_mapper = tRNAmap.tRNAmap()

from . import SSA_Soln
solution_obj = SSA_Soln

from . import ODE_Soln
ode_solution_obj = ODE_Soln

from . import TranslationOptimization
optimizer = TranslationOptimization
from ._version import __version__
from . import GenericMetaData



# from . import Core
# from . import Seq
# from .Core import *
# from .IO import *
# from .Seq import *
# from .Opt import *

# from .IntA import *
# from .Pfact import *
# from .Solvers import *



# -*- coding: utf-8 -*-
"""
Created on Thu May 20 11:34:48 2021

@author: William Raymond
"""

import warnings
import numpy as np
from . import SequenceManipMethods
smm = SequenceManipMethods.SequenceManipMethods


class DiffusionRateCalc():

    '''
    This is a dictionary object that handles everything
     dealing with tag sequences and AA tables
    '''

    def __init__(self):

        # g/mol or Daltons
        self.mw_table = {'a': 267.2413,          #adenosine
                        'AMP': 267.2413,
                        'g': 363.22,   #guanosine
                        'GMP': 363.22,
                        'u': 244.2014,    #uridine
                        'UMP': 244.2014,
                        'c': 243.217,   #cytidine
                        'CMP': 243.217,
                        '7mG': 363.22 + (12.017 + 3.021)*2 + 3*(94.9714),  #7 methyl G cap
                        'mean 80S': 2e6,
                        'human 80S': 4.3e6,  #https://pubmed.ncbi.nlm.nih.gov/25901680/
                        'tRNA_base': 25000, #approximation of 76 - 90 nt
                        'GFP': 26870,
                        'cy3':627.7,
                        'mCherry': 28000,
                        'Fab_blank': 48000,#approximation
                        'MS2CP': 13700,
                        'PCP': 16025,

            }

        # amino acid individual MWs
        self.aa_table_mw = {'A': 75,
                            'R':175,
                            'N': 132,
                            'D':133,
                            'B': 133,
                            'C': 121,
                            'Q':146,
                            'E': 147,
                            'Z':147,
                            'G':75,
                            'H':155,
                            'I':131,
                            'L':131,
                            'K':146,
                            'M':149,
                            'F':165,
                            'P':115,
                            'T':119,
                            'W':204,
                            'Y':181,
                            'V':117,
                            'S':105,
                            '*':0}



    def calculate_rna_strand_base_mw(self, seq, n_loops=0,
                                     fluorophore='GFP', coat_protein='MS2CP'):
        '''
        Return the base molecular weight (no ribosomes) of a strand of
        RNA with Fluorophore FAB loops

        Parameters
        ----------
        seq : str
            sequence string of the rna construct.
        n_loops : int, optional
            number of fluorophore tagging loops in the 3'UTR. The default is 0.
        fluorophore : str, optional
            type of fluorophore GFP or mCHerry. The default is 'GFP'.
        coat_protein : TYPE, optional
            type of coat protein for the MS2 or PP7 loops.
            The default is 'MS2CP'.
        Returns
        -------
        rna_weight : float
            Molecular weight in daltons of the construct.

        '''
        if isinstance(fluorophore, str):
            fluorphore_weight = self.mw_table[fluorophore]
        else:
            fluorphore_weight = fluorophore

        if isinstance(coat_protein, str):
            coat_weight = self.mw_table[coat_protein]
        else:
            coat_weight = coat_protein

        seq = seq.lower().replace('t','u')
        rna_weight = np.sum([self.mw_table[x.lower()] for x in seq])
        rna_weight += self.mw_table['7mG']  #add weight of the cap
        rna_weight += fluorphore_weight*n_loops #add weight of MS2 or PP7 loops
        rna_weight += coat_weight*n_loops  #add weight MS2CP or PCP

        return rna_weight



    def calculate_diffusion_constant(self, molecular_weight,
                                          viscosity=4.4e-2,
                                          tempC=37,
                                          r_fun='Default'):
        '''


        Parameters
        ----------
        molecular_weight : float or ndarray of floats
            Molecular weight to convert to diffusion constant (over time or
            singular value).
        viscosity : float, optional
            viscosity in pascal seconds. The default is 4.4e-2 Pascal seconds.
        tempC : float, optional
            temperature in Celcius. The default is 37.
        r_fun : lambda, optional
            function to convert MW to hydraulic radius.
            The default is Rh = MW^(1/3).

        Returns
        -------
        diffusion_coeff : float or ndarray of floats
            Diffusion coeff of the mrna strand over time.

        '''


        #reasonable range of diffusion: .6- 0.04 um^2/s

        #https://pubs.acs.org/doi/10.1021/nl2008218
        #https://pubs.acs.org/doi/10.1021/acs.jpclett.0c01748
        #https://www.ncbi.nlm.nih.gov/pmc/articles/PMC1831683/
        if r_fun == 'Default':
             ## approximation for now  #meters
            radius = (molecular_weight**.33)*1e-9
        else:
            radius = r_fun(molecular_weight)

        kbT = (tempC + 273.15)*1.380649e-23 # joules

        diffusion_coeff = kbT/ (6*np.pi * viscosity*radius)   #meters^2/s
        #Dr = kbT/ (8*np.pi * viscosity * radius**3)
        diffusion_coeff = diffusion_coeff*1e12 #convert to um^2/s
        return diffusion_coeff

    def calculate_single_rib_mw(self, aa_seq, probe_loc,
                                ribosome='human 80S', fluorophore='GFP',
                                fab='Fab_blank'):
        '''
        calculate the change in molecular weight for a single ribosome as it
        moves along codon position.

        Parameters
        ----------
        aa_seq : str
            Amino Acid sequence of the protein being translated.
        probe_loc : 2d numpy array
            probe locations, binary 0,1 for locations of probes.
        ribosome : str or float, optional
            key for the .mw_table or a molecular weight value in daltons.
            The default is 'human 80S'.
        fluorophore : str or float, optional
            key for the .mw_table or a molecular weight value in daltons.
            The default is 'GFP'.
        fab : TYPE, optional
            key for the .mw_table or a molecular weight value in daltons.
            The default is 'Fab_blank'.

        Returns
        -------
        weight_vec : 1d Numpy array
            molecular weight over position for a ribosome.

        '''

        if isinstance(ribosome, str):
            base_rib = self.mw_table[ribosome]
        else:
            base_rib = ribosome

        if isinstance(fab, str):
            fab_weight = self.mw_table['Fab_blank']
        else:
            fab_weight = fab

        if isinstance(fluorophore, str):
            fluorophore_weight = self.mw_table[fluorophore]
            fluorophore_weight = [fluorophore_weight]*probe_loc.shape[0]
        else:
            fluorophore_weight = fluorophore
            if not isinstance(fluorophore_weight, list):
                fluorophore_weight = [fluorophore_weight]

            if len(fluorophore_weight) >= probe_loc.shape[0]:
                fluorophore_weight = fluorophore_weight[:probe_loc.shape[0]]
            else:
                msg = 'warning: Provided probe location vector implies there'\
                      ' is more than one color, however only one fluorophore'\
                      ' was provided. Defaulting to repeating the one '\
                      'fluorophore MW multiple times. '
                warnings.warn(msg)
                fluorophore_weight = [fluorophore_weight[0]]*probe_loc.shape[0]

        probe_mw = probe_loc * np.array([np.sum(fluorophore_weight)
                                         + fab_weight])


        aa_mw = np.array([self.aa_table_mw[x] for x in aa_seq])
        weight_vec = base_rib + np.cumsum( probe_mw + aa_mw )
        return weight_vec

    def mw_over_time(self, ribosome_position_tensor, nt_seq, probe_loc,
                     base_mw,
                     ribosome='human 80S',
                     fluorophore='GFP',
                     fab='Fab_blank'):
        '''
        Calculate molecular weight over time for a ribosome position trajectory

        Parameters
        ----------
        X : 3d numpy array
            ribosome position tensor of size:
                [ntraj, n_time_points, n_ribs (at a time)].
        nt_seq : TYPE
            nucleotide sequence of the mRNA species.
        probe_loc : 2d Ndarray of probe locations
            binary vector describing where probe locations are.
        base_mw : float
            base molecular weight of the tagged RNA with no ribosomes actively
            translating.
        ribosome : str or float, optional
            key for the .mw_table or a molecular weight value in daltons.
            The default is 'human 80S'.
        fluorophore : str or float, optional
            key for the .mw_table or a molecular weight value in daltons.
            The default is 'GFP'.
        fab : TYPE, optional
            key for the .mw_table or a molecular weight value in daltons.
            The default is 'Fab_blank'.


        Returns
        -------
        1D numpy array
            molecular weight over time for a translating mRNA.

        '''

        if isinstance(fluorophore, str):
            fluorophore = self.mw_table[fluorophore]

        if isinstance(ribosome, str):
            ribosome = self.mw_table[ribosome]

        if isinstance(fab, str):
            fab = self.mw_table['Fab_blank']

        aa_seq = smm().nt2aa(nt_seq)
        weight_vec = self.calculate_single_rib_mw(aa_seq, probe_loc,
                                                  ribosome=ribosome,
                                                  fluorophore=fluorophore,
                                                  fab=fab)

        mw_per_pos = np.zeros(ribosome_position_tensor.shape[:-1])

        for i in range(ribosome_position_tensor.shape[0]):
            mw_per_pos[i,:] = np.sum(weight_vec[
                ribosome_position_tensor[i,:,:]], axis=1)

        return mw_per_pos + base_mw



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 18:13:32 2020

@author: willi
"""

import matplotlib.pyplot as plt
from matplotlib import cm
import matplotlib.patches as mpatches
import numpy as np
from dna_features_viewer import GraphicFeature, GraphicRecord

from . import SequenceManipMethods
from . import PropensityFactory, ProbeVectorFactory
from . import CodonDictionaries
PropensityFactory = PropensityFactory.PropensityFactory
ProbeVectorFactory = ProbeVectorFactory.ProbeVectorFactory
CodonDictionaries = CodonDictionaries.CodonDictionaries


class poi():
    '''

    Attributes
    ----------

    aa_seq: str
        amino acid sequence
    nt_seq: str
        nucleotide sequence
    gene_length: int
        length of the non-tag region if applicable
    tag_length: int
        length of the tag region if applicable
    name: str
        the name of the transcript
    tag_types: list
        list of tag types
    tag_epitopes: dict
        tag epitope dictionary: keys are the tag name, values are the tag nt seq
    ki: float
        initation rate
    ke_mu: float
        average elongation rate
    kt: float
        termination rate

    '''
    def __init__(self):
        self.aa_seq = ''    #amino sequence
        self.nt_seq = ''    #nucleotide sequence
        self.gene_length = 0   #length of the gene
        self.tag_length = 0   #length of the tags
        self.total_length = 0  #total length of the full amino acid sequence
        self.name = ''         #name of the gene
        self.tag_types = []
        self.tag_epitopes = {}  #type of tags and epitope lists per tag
        self.ki = .03
        self.ke_mu = 10
        self.kt = 10

        self.multiframe_epitopes = []
        self.multiframe_nt_seq = []
        self.multiframe_aa_seq = []
        self._colors = ['#fa8174', '#b3de69', '#bc82bd','#ccebc4',
                        '#ffed6f','#81b1d2']


    @property
    def CAI(self):
        '''
        Geometric mean of the weight of to each codon in a transcript.

        ..[1] https://en.wikipedia.org/wiki/Codon_Adaptation_Index

        Returns
        -------
        CAI : Float
            Codon Adaptation Index. - a 0-1 measure of codon bias.



        '''

        _, CAI, _ = SequenceManipMethods().codon_usage(self.nt_seq)
        return CAI

    @property
    def codon_sensitivity(self):
        '''
        returns the percentage usage of each codon within the transcript
        times its senstitivity rate from cdict

        %codon in transcript * sensitivity

        sensitivity is defined as the ratio of the fastest synonomous codon
        rate / slowest synonomous codon rate

        Returns
        -------
        codon_sensitivity : list of floats


        '''

        cs, _, _ = SequenceManipMethods().codon_usage(self.nt_seq)
        return cs



    @property
    def codons(self):
        '''
        A list of all codon strings

        Returns
        -------
        list
            list of all codons in the protein sequence.

        '''
        codons = self.nt_seq.upper()
        return [codons[i:i+3] for i in range(0, len(codons), 3)]

    @property
    def ktrna_id(self):
        return PropensityFactory().get_trna_ids(self.nt_seq)

    @property
    def kelong(self):
        '''
        quick method to get the default elongation rates for a transcript from its current ke_mu

        Returns
        -------
        list
            elongation rates for the protein object, calculated from the default gene copy numbers

        '''
        return PropensityFactory().get_k(self.nt_seq, self.ki, self.ke_mu,
                                         self.kt)[1:-1]

    @property
    def probe_vec(self):
        '''
        probe vector describes the transformation of ribosome position to intensity value

        For example, a transcript of length 5 with the probe locations

        [0,1,0,1,0]

        would have the following probe vector

        [0,1,1,2,2]

        Returns
        -------
        probe_vec : ndarray
            cumulative sum vector for Ncolor x L of the transcript.

        '''
        pv = np.zeros((len(list(self.tag_epitopes)), self.total_length))
        for i in range(len(list(self.tag_epitopes))):
            pv[i, [self.tag_epitopes[list(self.tag_epitopes.keys())[i]]]] = 1
        pv = np.cumsum(pv, axis=1)
        return pv.astype(np.int32)

    @property
    def probe_loc(self):
        '''
        Probe location vector, describes the indexes where the epitope locations are

        Returns
        -------
        probe_loc : ndarray
            binary vector of epitope locations over Ncolor x L of the transcript.

        '''
        pv = np.zeros((len(list(self.tag_epitopes)), self.total_length), dtype = np.int32)
        for i in range(len(list(self.tag_epitopes))):
            pv[i, [self.tag_epitopes[list(self.tag_epitopes.keys())[i]]]] = 1
        return pv


    @property
    def all_k(self):
        '''
        All propensities for simulation

        Returns
        -------
        all propensities vector : list
            a list of all propensities including kinitiation and ktermination (ki, kt)

        '''
        return PropensityFactory().get_k(self.nt_seq, self.ki, self.ke_mu, self.kt)


    def get_binned_vectors(self, nbins, strategy='intellegent',
                           min_binsize=3):
        if strategy == 'intellegent':
            inds = PropensityFactory().intellegent_bin(self.probe_loc,
                                                       nbins,
                                                       min_bin=min_binsize)

        if strategy == 'even':
            inds = PropensityFactory().even_bin(len(self.kelong), nbins)

        bin_k = PropensityFactory().bin_k(self.kelong, inds)
        probeloc_binned, probevec_binned = ProbeVectorFactory().bin_probe_vecs(self.probe_loc, inds)

        return inds, probeloc_binned, probevec_binned, bin_k


    def generate_3frame_tags(self):
        '''
        Generates 3 open reading frame tags for model building on the POI obj

        * multiframe_epitopes
        * multiframe_nt_seq
        * multiframe_aa_seq

        Returns
        -------
        None.

        '''

        #multiframe_epitopes = {}
        codons_seq = ''
        for n in range(3):
            if n == 0:
                codons_seq = codons_seq + self.nt_seq
                self.multiframe_nt_seq.append(self.nt_seq)
            else:
                codons_seq = codons_seq + self.nt_seq[n:-(3-n)]
                self.multiframe_nt_seq.append(self.nt_seq[n:-(3-n)])

        cd = CodonDictionaries()
        smm = SequenceManipMethods.SequenceManipMethods()
        #codons_aa_seq = smm.nt2aa(codons_seq)

        for frame in self.multiframe_nt_seq:
            self.multiframe_aa_seq.append(smm.nt2aa(frame))

        for frame in self.multiframe_aa_seq:
            multiframe_epitopes = {}
            for tag in cd.tag_dict.keys():

                if cd.tag_dict[tag] in frame:

                    #tag_detected = True

                    epi = smm.get_tag_loc(frame, cd.tag_dict[tag])

                    multiframe_epitopes[tag] = epi

            self.multiframe_epitopes.append(multiframe_epitopes)


    def save(self):
        '''
        TODO
        '''
        x = 1
    def load(self):
        '''
        TODO
        '''
        x = 1


    def visualize_mrna_strand(self, dpi=120, cmap='viridis',
                              additional_features=[]):
        features = [
            GraphicFeature(start=0, end=self.tag_length,
                                   color=self._colors[0], label='Tag'),
            GraphicFeature(start=self.tag_length, end=self.total_length,
                                   color=self._colors[1], label='Protein'),
                    ]
        
        probe = self.probe_loc
        cmap = cm.get_cmap(cmap)
        color = np.where(probe == 1)[0]
        location = np.where(probe == 1)[1]
        
        ncolors = probe.shape[0]
        colors = cmap(np.linspace(.01, .95, ncolors))
        colorlabels = ['Color %d'% i for i in range(ncolors)]
        for c, loc in zip(color, location):
            features = features + [
                GraphicFeature(start=loc, end=loc+2,
                               color=colors[c], linecolor=colors[c]),
                ]
    
        features += additional_features
        record = GraphicRecord(sequence_length=self.total_length,
                               features=features)
        
        fig, ax = plt.subplots(1, dpi=dpi)
        
        for c in range(ncolors):
            ax.plot([0, 0], [0,0], color=colors[c])  #fix the legend colors

        
        colorlabels = ['Color %d'% i for i in range(ncolors)]
        ax,_ = record.plot(figure_width=6, ax=ax)

        ax.axes.legend(colorlabels, loc=7)
        ax.text(0, 5, 'Transcript Name: %s' % self.name)
        ax.text(0, 4, 'Total Length: %d codons' % self.total_length)
        ax.text(0, 3, 'Seq: %s ...' % self.aa_seq[:10])
        fig.show()


    def visualize_probe(self, colors=None, dpi=120, cmap='viridis'):
        '''
        method to visualize the transcript

        Parameters
        ----------
        colors : list, optional
            Ncolor list of colors to use for the tag colors. The default is cmap gist rainbow.

        Returns
        -------
        None.

        '''
        probe = self.probe_loc
        fig, ax = plt.subplots(1, dpi=dpi)
        N = len(self.kelong)
        ncolors = probe.shape[0]

        cmap = cm.get_cmap(cmap)
        if colors == None:
            colors = cmap(np.linspace(.01, .95, ncolors))

        rectangle = mpatches.Rectangle((0, .1), N, .8,
                                       linewidth=1, edgecolor='k',
                                       facecolor='gray')

        ax.add_patch(rectangle)

        color = np.where(probe == 1)[0]
        location = np.where(probe == 1)[1]

        colorlabels = ['Color %d'% i for i in range(ncolors)]

        for c in range(ncolors):
            ax.plot([-10, -10], [-4, -4], color=colors[c])  #fix the legend colors


        for c,loc in zip(color, location):
            ax.plot([loc, loc], [.1, .9], color=colors[c])

        ax.set_ylim([-.1, 10])
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.set_xlabel('codon')
        ax.axes.legend(colorlabels, loc=7)
        ax.axes.get_yaxis().set_visible(False)
        
        
        ax.text(0, 5, 'Transcript Name: %s' % self.name)
        ax.text(0, 4, 'Total Length: %d codons' % self.total_length)
        ax.text(0, 3, 'Seq: %s ...' % self.aa_seq[:10])

        fig.show()



import numpy as np
import scipy.sparse.linalg as spl

def expv(t,A,v,tol=1.0e-7,m=30, u = 0):
    '''
    a python version of expv from 
    roger sidje's expokit. A should be sparse.
    
    https://www.maths.uq.edu.au/expokit/paper.pdf
    
    Produces the analytical solution to the Homogenous ODE solutoin
    
    dw/dt = A* w(t) with w(0) = v
    dw/dt = A* w(t) + u with w(0) = v TODO add this <--
    
    
    
    Parameters
    ----------
    t : timepoint to solve
        time vector to solve the ODE over
    A : ndarray (states x reactions)
        The state transition matrix
    v : ndarray (states x 1)
        initial state of the system

    Returns
    -------
    w - Solution
    err - Error in approximation
    hump - bound approximation

    '''
    n,n = A.shape 
    if n<m:
        m = n
    anorm = spl.norm(A,np.inf) 

    # set some default parameters
    mxrej = 10.0;  btol  = 1.0e-7; 
    gamma = 0.9; delta = 1.2; 
    mb    = m; t_out   = abs(t);
    nstep = 0.0; t_new   = 0.0;
    t_now = 0.0; s_error = 0.0;
    eps = np.finfo(float).eps 
    rndoff= anorm*eps; 

    k1 = 2; xm = 1.0/m; normv = np.linalg.norm(v); beta = normv;
    fact = (((m+1)/np.exp(1))**(m+1))*np.sqrt(2*np.pi*(m+1));
    t_new = (1.0/anorm)*((fact*tol)/(4.0*beta*anorm))**xm;
    s = 10**(np.floor(np.log10(t_new))-1); t_new = np.ceil(float(t_new)/s)*s; 
    sgn = np.sign(t); nstep = 0;   
    w = v
    hump  = normv
    while t_now < t_out:
        nstep += 1
        t_step = np.min([t_out-t_now,t_new])
        V = np.zeros((n,m+1))
        H = np.zeros((m+2,m+2))
        V[:,0] = np.ravel((1.0/beta)*w)
        for j in range(m):
            p = A.dot(V[:,j])
            for i in range(j+1): 
                H[i,j] = np.dot(V[:,i].T,p)
                p = p-H[i,j]*V[:,i]
            s = np.linalg.norm(p)
            if s<btol: 
                k1 = 0.0
                mb = j+1
                t_step = t_out - t_now
                break                
            H[j+1,j] = s
            V[:,j+1] = (1/s)*p
        if k1 != 0:
            H[m+1,m] = 1
            avnorm = np.linalg.norm(A.dot(V[:,m])) # check index.
        ireject = 0
        while ireject<=mxrej:
            mx = int(mb+k1)
            F = spl.expm(sgn*t_step*H[:mx,:mx])
            if k1 == 0:
                err_loc = btol
                break
            else:
                phi1 = np.abs(beta*F[m,1])
                phi2 = np.abs(beta*F[m+1,1] * avnorm)
                if phi1 > 10*phi2:
                    err_loc = phi2
                    xm = 1.0/float(m)
                elif phi1>phi2:
                    err_loc = (phi1*phi2)/(phi1-phi2)
                    xm = 1.0/float(m)
                else:
                    err_loc = phi1
                    xm = 1.0 / float(m-1)
                
            if err_loc <= delta * t_step * tol:
                break
            else:
                t_step = gamma * t_step * (t_step * tol/err_loc)**xm
                s = 10**(np.floor(np.log10(t_step))-1)
                t_step = np.ceil(t_step/s)*s
                if ireject == mxrej:
                    print('The requested tolerance is too high.')
                ireject += 1
      
        mx = int(mb) + int(np.max( [0,k1-1]))                    
        w = np.dot(V[:,:mx],beta*F[:mx,0])
        beta = np.linalg.norm(w)
        hump = np.max([hump,beta])
        t_now += t_step
        t_new = gamma * t_step * (t_step*tol/float(err_loc))**xm
        s = 10**(np.floor(np.log10(t_new))-1)
        t_new = np.ceil(t_new/float(s))*s
        err_loc = np.max([err_loc,rndoff])
        s_error = s_error + err_loc
    err = s_error
    hump = hump / float(normv)
    return w, err, hump
        



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 24 11:44:48 2018

@author: william
"""

import setuptools


import re
VERSIONFILE="rsnapsim/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rsnapsim", 
    version=verstr,
    author="Dr. Luis Aguilera, William Raymond, Dr. Brian Munsky",
    author_email="wsraymon@rams.colostate.edu",
    description="A package for mRNA sequence translation stochastic simulations",
    long_description=long_description,
    long_description_content_type="text/markdown",
    #url="githublink",
    packages=setuptools.find_packages(exclude=('ssa_cpp.*','test.*','models.*','trna_ssa.*','ssa_cpp')),
    include_package_data=True,
    package_data = {"rsnapsim":['*.h', '*.cpp','*.pyx','license.txt']},
    install_requires = ['BioPython','numpy','scipy','cython','matplotlib','pandas', 'SnapGeneFileReader','dna_features_viewer'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)